"""Class for the input data of the fine-mapping analysis."""

import logging
import os
from typing import Any, Dict, List, Optional

import numpy as np
import pandas as pd

from credtools.constants import ColName
from credtools.ldmatrix import LDMatrix, load_ld
from credtools.sumstats import load_sumstats

logger = logging.getLogger("Locus")


class Locus:
    """
    Locus class to represent a genomic locus with associated summary statistics and linkage disequilibrium (LD) matrix.

    Parameters
    ----------
    popu : str
        Population code. e.g. "EUR". Choose from ["AFR", "AMR", "EAS", "EUR", "SAS"].
    cohort : str
        Cohort name.
    sample_size : int
        Sample size.
    sumstats : pd.DataFrame
        Summary statistics DataFrame.
    ld : Optional[LDMatrix], optional
        LD matrix, by default None.
    if_intersect : bool, optional
        Whether to intersect the LD matrix and summary statistics file, by default False.

    Attributes
    ----------
    original_sumstats : pd.DataFrame
        The original summary statistics file.
    sumstats : pd.DataFrame
        The processed summary statistics file.
    ld : LDMatrix
        The LD matrix object.
    chrom : int
        Chromosome.
    start : int
        Start position of the locus.
    end : int
        End position of the locus.
    n_snps : int
        Number of SNPs in the locus.
    prefix : str
        The prefix combining population and cohort.
    locus_id : str
        Unique identifier for the locus.
    is_matched : bool
        Whether the LD matrix and summary statistics file are matched.
    lambda_s : Optional[float]
        The estimated lambda_s parameter from estimate_s_rss function, None if not calculated.

    Notes
    -----
    If no LD matrix is provided, only ABF method can be used for fine-mapping.
    """

    def __init__(
        self,
        popu: str,
        cohort: str,
        sample_size: int,
        sumstats: pd.DataFrame,
        locus_start: int,
        locus_end: int,
        ld: Optional[LDMatrix] = None,
        if_intersect: bool = False,
    ) -> None:
        """
        Initialize the Locus object.

        Parameters
        ----------
        popu : str
            Population code. e.g. "EUR". Choose from ["AFR", "AMR", "EAS", "EUR", "SAS"].
        cohort : str
            Cohort name.
        sample_size : int
            Sample size.
        sumstats : pd.DataFrame
            Summary statistics DataFrame.
        locus_start : int
            Fixed start position for the locus.
        locus_end : int
            Fixed end position for the locus.
        ld : Optional[LDMatrix], optional
            LD matrix, by default None.
        if_intersect : bool, optional
            Whether to intersect the LD matrix and summary statistics file, by default False.

        Warnings
        --------
        If no LD matrix is provided, a warning is logged that only ABF method can be used.
        """
        self.sumstats = sumstats
        self._original_sumstats = self.sumstats.copy()
        self._popu = popu
        self._cohort = cohort
        self._sample_size = sample_size
        self._locus_start = locus_start
        self._locus_end = locus_end
        self.lambda_s = None
        if ld:
            self.ld = ld
            if if_intersect:
                inters = intersect_sumstat_ld(self)
                self.sumstats = inters.sumstats
                self.ld = inters.ld
        else:
            logger.warning("LD matrix and map file not found. Can only run ABF method.")
            self.ld = LDMatrix(pd.DataFrame(), np.array([]))

    @property
    def original_sumstats(self) -> pd.DataFrame:
        """Get the original sumstats file."""
        return self._original_sumstats

    @property
    def popu(self) -> str:
        """Get the population code."""
        return self._popu

    @property
    def cohort(self) -> str:
        """Get the cohort name."""
        return self._cohort

    @property
    def sample_size(self) -> int:
        """Get the sample size."""
        return self._sample_size

    @property
    def chrom(self) -> int:
        """Get the chromosome."""
        return self.sumstats[ColName.CHR].iloc[0]

    @property
    def start(self) -> int:
        """Get the start position."""
        return self._locus_start

    @property
    def end(self) -> int:
        """Get the end position."""
        return self._locus_end

    @property
    def n_snps(self) -> int:
        """Get the number of SNPs."""
        return len(self.sumstats)

    @property
    def prefix(self) -> str:
        """Get the prefix of the locus."""
        import hashlib

        # Check if this is a meta-analysis result (contains '+' in cohort)
        if "+" in self.cohort:
            # Create a hash of the cohort part
            cohort_hash = hashlib.md5(self.cohort.encode()).hexdigest()[:8]
            # Count number of cohorts for reference
            num_cohorts = len(self.cohort.split("+"))
            return f"{self.popu}_meta{num_cohorts}cohorts_{cohort_hash}"

        # For single cohort, use original format
        return f"{self.popu}_{self.cohort}"

    @property
    def locus_id(self) -> str:
        """Get the locus ID."""
        return f"{self.popu}_{self.cohort}_chr{self.chrom}:{self.start}-{self.end}"

    @property
    def is_matched(self) -> bool:
        """Check if the LD matrix and sumstats file are matched."""
        # check the order of SNPID in the LD matrix and the sumstats file are the exact same
        if self.ld is None:
            return False
        return self.ld.map[ColName.SNPID].equals(self.sumstats[ColName.SNPID])

    def __repr__(self) -> str:
        """
        Return a string representation of the Locus object.

        Returns
        -------
        str
            String representation of the Locus object.
        """
        return f"Locus(popu={self.popu}, cohort={self.cohort}, sample_size={self.sample_size}, chr={self.chrom}, start={self.start}, end={self.end}, sumstats={self.sumstats.shape}, ld={self.ld.r.shape})"

    def copy(self) -> "Locus":
        """
        Copy the Locus object.

        Returns
        -------
        Locus
            A copy of the Locus object.
        """
        new_locus = Locus(
            self.popu,
            self.cohort,
            self.sample_size,
            self.sumstats.copy(),
            self._locus_start,
            self._locus_end,
            self.ld.copy(),
            if_intersect=False,
        )
        new_locus.lambda_s = self.lambda_s
        return new_locus


class LocusSet:
    """
    LocusSet class to represent a set of genomic loci.

    Parameters
    ----------
    loci : List[Locus]
        List of Locus objects.

    Attributes
    ----------
    loci : List[Locus]
        List of Locus objects.
    n_loci : int
        Number of loci.
    chrom : int
        Chromosome number.
    start : int
        Start position of the locus.
    end : int
        End position of the locus.
    locus_id : str
        Unique identifier for the locus.

    Raises
    ------
    ValueError
        If the chromosomes of the loci are not the same.
    """

    def __init__(self, loci: List[Locus]) -> None:
        """
        Initialize the LocusSet object.

        Parameters
        ----------
        loci : List[Locus]
            List of Locus objects.
        """
        self.loci = loci

    @property
    def n_loci(self) -> int:
        """Get the number of loci."""
        return len(self.loci)

    @property
    def chrom(self) -> int:
        """
        Get the chromosome.

        Returns
        -------
        int
            Chromosome number.

        Raises
        ------
        ValueError
            If the chromosomes of the loci are not the same.
        """
        chrom_list = [locus.chrom for locus in self.loci]
        if len(set(chrom_list)) > 1:
            raise ValueError("The chromosomes of the loci are not the same.")
        return chrom_list[0]

    @property
    def start(self) -> int:
        """Get the start position."""
        # All loci should have the same fixed start
        starts = [locus._locus_start for locus in self.loci]
        if not all(s == starts[0] for s in starts):
            raise ValueError("All loci in LocusSet must have the same start position")
        return starts[0]

    @property
    def end(self) -> int:
        """Get the end position."""
        # All loci should have the same fixed end
        ends = [locus._locus_end for locus in self.loci]
        if not all(e == ends[0] for e in ends):
            raise ValueError("All loci in LocusSet must have the same end position")
        return ends[0]

    @property
    def locus_id(self) -> str:
        """Get the locus ID."""
        return f"{self.chrom}:{self.start}-{self.end}"

    def __repr__(self) -> str:
        """
        Return a string representation of the LocusSet object.

        Returns
        -------
        str
            String representation of the LocusSet object.
        """
        return (
            f"LocusSet(\n n_loci={len(self.loci)}, chrom={self.chrom}, start={self.start}, end={self.end}, locus_id={self.locus_id} \n"
            + "\n".join([locus.__repr__() for locus in self.loci])
            + "\n"
            + ")"
        )

    def copy(self) -> "LocusSet":
        """
        Copy the LocusSet object.

        Returns
        -------
        LocusSet
            A copy of the LocusSet object.
        """
        return LocusSet([locus.copy() for locus in self.loci])


def intersect_sumstat_ld(locus: Locus) -> Locus:
    """
    Intersect the Variant IDs in the LD matrix and the sumstats file.

    Parameters
    ----------
    locus : Locus
        Locus object containing LD matrix and summary statistics.

    Returns
    -------
    Locus
        Locus object containing the intersected LD matrix and sumstats file.

    Raises
    ------
    ValueError
        If LD matrix not found or no common Variant IDs found between the LD matrix and the sumstats file.

    Warnings
    --------
    If only a few common Variant IDs are found (≤ 10), a warning is logged.

    Notes
    -----
    This function performs the following operations:

    1. Checks if LD matrix and summary statistics are already matched
    2. Finds common SNP IDs between LD matrix and summary statistics
    3. Subsets both datasets to common variants
    4. Reorders data to maintain consistency
    5. Returns a new Locus object with intersected data
    """
    if locus.ld is None:
        raise ValueError("LD matrix not found.")
    if locus.is_matched:
        logger.info("The LD matrix and sumstats file are matched.")
        return locus
    ldmap = locus.ld.map.copy()
    r = locus.ld.r.copy()
    sumstats = locus.sumstats.copy()
    sumstats = sumstats.sort_values([ColName.CHR, ColName.BP], ignore_index=True)
    intersec_sumstats = sumstats[
        sumstats[ColName.SNPID].isin(ldmap[ColName.SNPID])
    ].copy()
    intersec_variants = intersec_sumstats[ColName.SNPID].to_numpy()
    if len(intersec_variants) == 0:
        raise ValueError(
            f"No common Variant IDs found between the LD matrix and the sumstats file for locus {locus.locus_id}."
        )
    elif len(intersec_variants) <= 10:
        logger.warning(
            f"Only a few common Variant IDs found between the LD matrix and the sumstats file(<= 10) for locus {locus.locus_id}."
        )
    ldmap["idx"] = ldmap.index
    ldmap.set_index(ColName.SNPID, inplace=True, drop=False)
    ldmap = ldmap.loc[intersec_variants].copy()
    intersec_index = ldmap["idx"].to_numpy()
    r = r[intersec_index, :][:, intersec_index]
    intersec_sumstats.reset_index(drop=True, inplace=True)
    ldmap.drop("idx", axis=1, inplace=True)
    ldmap = ldmap.reset_index(drop=True)
    intersec_ld = LDMatrix(ldmap, r)
    logger.info(
        "Intersected the Variant IDs in the LD matrix and the sumstats file. "
        f"Number of common Variant IDs: {len(intersec_index)}"
    )
    return Locus(
        locus.popu,
        locus.cohort,
        locus.sample_size,
        intersec_sumstats,
        locus._locus_start,
        locus._locus_end,
        intersec_ld,
        if_intersect=False,
    )


def check_loci_info(loci_info: pd.DataFrame) -> pd.DataFrame:
    """
    Check and validate loci information DataFrame.

    Parameters
    ----------
    loci_info : pd.DataFrame
        DataFrame containing loci information.

    Returns
    -------
    pd.DataFrame
        Validated and type-corrected loci_info DataFrame.

    Raises
    ------
    ValueError
        If required columns are missing, data types are incorrect,
        or locus_id/boundary consistency checks fail.

    Notes
    -----
    This function performs the following checks:
    1. Ensures all required columns are present
    2. Validates and converts data types
    3. Checks that loci with same locus_id have same chr, start, end
    4. Validates chromosome, start, and end values
    """
    loci_info = loci_info.copy()

    # Check for required columns
    required_cols = [
        "prefix",
        "popu",
        "cohort",
        "sample_size",
        "chr",
        "start",
        "end",
        "locus_id",
    ]
    missing_cols = [col for col in required_cols if col not in loci_info.columns]
    if missing_cols:
        raise ValueError(f"Missing required columns: {missing_cols}")

    # Type checking and conversion
    try:
        # Convert numeric columns
        loci_info["sample_size"] = loci_info["sample_size"].astype(int)
        loci_info["chr"] = loci_info["chr"].astype(int)
        loci_info["start"] = loci_info["start"].astype(int)
        loci_info["end"] = loci_info["end"].astype(int)

        # Ensure string columns are strings
        loci_info["prefix"] = loci_info["prefix"].astype(str)
        loci_info["popu"] = loci_info["popu"].astype(str)
        loci_info["cohort"] = loci_info["cohort"].astype(str)
        loci_info["locus_id"] = loci_info["locus_id"].astype(str)

    except (ValueError, TypeError) as e:
        raise ValueError(f"Failed to convert data types: {e}")

    # Validate values
    if (loci_info["sample_size"] <= 0).any():
        raise ValueError("Sample size must be positive")

    if (loci_info["chr"] <= 0).any() or (loci_info["chr"] > 25).any():
        raise ValueError("Chromosome must be between 1 and 25")

    if (loci_info["start"] <= 0).any():
        raise ValueError("Start position must be positive")

    if (loci_info["end"] <= loci_info["start"]).any():
        raise ValueError("End position must be greater than start position")

    # Check for duplicates in popu+cohort+locus_id combination
    if loci_info.duplicated(subset=["popu", "cohort", "locus_id"]).any():
        raise ValueError("Each popu+cohort+locus_id combination must be unique")

    # Check consistency: same locus_id must have same chr, start, end
    locus_boundaries = loci_info.groupby("locus_id")[["chr", "start", "end"]].nunique()
    inconsistent_loci = locus_boundaries[(locus_boundaries > 1).any(axis=1)]

    if not inconsistent_loci.empty:
        raise ValueError(
            f"Inconsistent boundaries for locus_id(s): {inconsistent_loci.index.tolist()}. "
            "Each locus_id must have consistent chr, start, end values across all rows."
        )

    return loci_info


def intersect_loci(list_loci: List[Locus]) -> List[Locus]:
    """
    Intersect the Variant IDs in the LD matrices and the sumstats files of a list of Locus objects.

    Parameters
    ----------
    list_loci : List[Locus]
        List of Locus objects.

    Returns
    -------
    List[Locus]
        List of Locus objects containing the intersected LD matrices and sumstats files.

    Raises
    ------
    NotImplementedError
        This function is not yet implemented.

    Notes
    -----
    This function is planned to intersect variant IDs across multiple loci
    to ensure consistent variant sets for multi-ancestry analysis.
    """
    raise NotImplementedError(
        "Intersect the Variant IDs in the LD matrices and the sumstats files of a list of Locus objects."
    )


def load_locus(
    prefix: str,
    popu: str,
    cohort: str,
    sample_size: int,
    locus_start: int,
    locus_end: int,
    if_intersect: bool = False,
    calculate_lambda_s: bool = False,
    **kwargs: Any,
) -> Locus:
    """
    Load the input data of the fine-mapping analysis.

    Parameters
    ----------
    prefix : str
        Prefix of the input files.
    popu : str
        Population of the input data.
    cohort : str
        Cohort of the input data.
    sample_size : int
        Sample size of the input data.
    locus_start : int
        Fixed start position for the locus.
    locus_end : int
        Fixed end position for the locus.
    if_intersect : bool, optional
        Whether to intersect the input data with the LD matrix, by default False.
    calculate_lambda_s : bool, optional
        Whether to calculate lambda_s parameter using estimate_s_rss function, by default False.
    **kwargs : Any
        Additional keyword arguments passed to loading functions.

    Returns
    -------
    Locus
        Locus object containing the input data.

    Raises
    ------
    ValueError
        If the required input files are not found.

    Notes
    -----
    The function looks for files with the following patterns:

    - Summary statistics: {prefix}.sumstat or {prefix}.sumstats.gz
    - LD matrix: {prefix}.ld or {prefix}.ld.npz
    - LD map: {prefix}.ldmap or {prefix}.ldmap.gz

    All files are required for proper functioning.

    Examples
    --------
    >>> locus = load_locus('EUR_study1', 'EUR', 'study1', 50000)
    >>> print(f"Loaded locus with {locus.n_snps} SNPs")
    Loaded locus with 10000 SNPs
    """
    if os.path.exists(f"{prefix}.sumstat"):
        sumstats_path = f"{prefix}.sumstat"
    elif os.path.exists(f"{prefix}.sumstats.gz"):
        sumstats_path = f"{prefix}.sumstats.gz"
    else:
        raise ValueError("Sumstats file not found.")

    sumstats = load_sumstats(sumstats_path, if_sort_alleles=True, **kwargs)
    if os.path.exists(f"{prefix}.ld"):
        ld_path = f"{prefix}.ld"
    elif os.path.exists(f"{prefix}.ld.npz"):
        ld_path = f"{prefix}.ld.npz"
    else:
        raise ValueError("LD matrix file not found.")
    if os.path.exists(f"{prefix}.ldmap"):
        ldmap_path = f"{prefix}.ldmap"
    elif os.path.exists(f"{prefix}.ldmap.gz"):
        ldmap_path = f"{prefix}.ldmap.gz"
    else:
        raise ValueError("LD map file not found.")
    ld = load_ld(ld_path, ldmap_path, if_sort_alleles=True, **kwargs)

    locus = Locus(
        popu,
        cohort,
        sample_size,
        sumstats,
        locus_start,
        locus_end,
        ld=ld,
        if_intersect=if_intersect,
    )

    if calculate_lambda_s:
        try:
            # Import here to avoid circular imports
            from credtools.qc import estimate_s_rss

            locus.lambda_s = estimate_s_rss(locus)
            logger.info(
                f"Calculated lambda_s for locus {locus.locus_id}: {locus.lambda_s}"
            )
        except Exception as e:
            logger.warning(
                f"Failed to calculate lambda_s for locus {locus.locus_id}: {e}"
            )
            locus.lambda_s = None

    return locus


def load_locus_set(
    locus_info: pd.DataFrame,
    if_intersect: bool = False,
    calculate_lambda_s: bool = False,
    **kwargs: Any,
) -> LocusSet:
    """
    Load the input data of the fine-mapping analysis for multiple loci.

    Parameters
    ----------
    locus_info : pd.DataFrame
        DataFrame containing the locus information with required columns:
        ['prefix', 'popu', 'cohort', 'sample_size', 'chr', 'start', 'end', 'locus_id'].
    if_intersect : bool, optional
        Whether to intersect the input data with the LD matrix, by default False.
    calculate_lambda_s : bool, optional
        Whether to calculate lambda_s parameter using estimate_s_rss function, by default False.
    **kwargs : Any
        Additional keyword arguments passed to load_locus function.

    Returns
    -------
    LocusSet
        LocusSet object containing the input data.

    Raises
    ------
    ValueError
        If required columns are missing or if the combination of popu and cohort is not unique.

    Notes
    -----
    The locus_info DataFrame must contain the following columns:

    - prefix: File prefix for each locus
    - popu: Population code
    - cohort: Cohort name
    - sample_size: Sample size for the cohort
    - chr: Chromosome number
    - start: Start position of the locus
    - end: End position of the locus
    - locus_id: Locus identifier

    All rows must have the same chr, start, end, locus_id values (representing the same locus).

    Examples
    --------
    >>> locus_info = pd.DataFrame({
    ...     'prefix': ['EUR_study1', 'ASN_study2'],
    ...     'popu': ['EUR', 'ASN'],
    ...     'cohort': ['study1', 'study2'],
    ...     'sample_size': [50000, 30000]
    ... })
    >>> locus_set = load_locus_set(locus_info)
    >>> print(f"Loaded {locus_set.n_loci} loci")
    Loaded 2 loci
    """
    # Check and validate the locus_info DataFrame
    locus_info = check_loci_info(locus_info)

    # Check that all rows have the same chr, start, end (same locus)
    if len(locus_info["chr"].unique()) > 1:
        raise ValueError("All rows must have the same chromosome")
    if len(locus_info["start"].unique()) > 1:
        raise ValueError("All rows must have the same start position")
    if len(locus_info["end"].unique()) > 1:
        raise ValueError("All rows must have the same end position")
    if len(locus_info["locus_id"].unique()) > 1:
        raise ValueError("All rows must have the same locus_id")

    # Additional check for load_locus_set: popu+cohort must be unique within this single locus
    if locus_info.duplicated(subset=["popu", "cohort"]).any():
        raise ValueError(
            "Each popu+cohort combination must be unique within a single locus"
        )

    loci = []
    for i, row in locus_info.iterrows():
        loci.append(
            load_locus(
                row["prefix"],
                row["popu"],
                row["cohort"],
                row["sample_size"],
                int(row["start"]),
                int(row["end"]),
                if_intersect,
                calculate_lambda_s,
                **kwargs,
            )
        )
    return LocusSet(loci)



"""
Internal munging module for credtools.

Adapted from smunger (https://github.com/Jianhua-Wang/smunger)
Original author: Jianhua Wang
License: MIT

This module provides GWAS summary statistics munging functionality
without requiring external dependencies.

Main Functions
--------------
munge : Main data munging and cleaning function
read_config : Read configuration from file
inspect_headers : Examine file headers
create_config_template : Generate column mapping configuration
"""

import json
import logging
import os
from typing import Any, Dict, List, Optional, Union

import pandas as pd

from .constants import ColName
from .core import make_SNPID_unique, munge
from .headers import (
    apply_header_mapping,
    create_config_template,
    inspect_headers,
    map_headers_automatic,
    validate_required_columns,
)

logger = logging.getLogger("Munging")

# Main API functions (compatible with smunger)
__all__ = [
    "munge",
    "read_config",
    "inspect_headers",
    "create_config_template",
    "make_SNPID_unique",
    "load_and_munge",
]


def read_config(config_path: str) -> Dict[str, Any]:
    """
    Read configuration from JSON file.

    Parameters
    ----------
    config_path : str
        Path to JSON configuration file.

    Returns
    -------
    Dict[str, Any]
        Configuration dictionary.

    Raises
    ------
    FileNotFoundError
        If configuration file doesn't exist.
    ValueError
        If configuration file is invalid JSON.
    """
    if not os.path.exists(config_path):
        raise FileNotFoundError(f"Configuration file not found: {config_path}")

    try:
        with open(config_path, "r") as f:
            config = json.load(f)

        logger.info(f"Loaded configuration from {config_path}")
        return config

    except json.JSONDecodeError as e:
        raise ValueError(f"Invalid JSON in configuration file: {e}")


def load_and_munge(
    file_path: str,
    config: Optional[Dict[str, Any]] = None,
    sep: Optional[str] = None,
    **kwargs,
) -> pd.DataFrame:
    """
    Load file and perform munging in one step.

    Parameters
    ----------
    file_path : str
        Path to input file.
    config : Dict[str, Any], optional
        Configuration dictionary with column mappings.
    sep : str, optional
        Column separator. Auto-detected if None.
    **kwargs
        Additional arguments for pandas.read_csv.

    Returns
    -------
    pd.DataFrame
        Munged DataFrame.
    """
    logger.info(f"Loading and munging file: {file_path}")

    # Auto-detect separator if not provided
    if sep is None:
        from .headers import _detect_separator

        sep = _detect_separator(file_path)

    # Load data
    compression = "gzip" if file_path.endswith(".gz") else None
    df = pd.read_csv(file_path, sep=sep, compression=compression, **kwargs)

    logger.info(f"Loaded {len(df)} rows with {len(df.columns)} columns")

    # Apply column mapping if provided in config
    if config and "column_mapping" in config:
        df = apply_header_mapping(df, config["column_mapping"])
    else:
        # Auto-map headers
        headers = df.columns.tolist()
        mapping = map_headers_automatic(headers)
        df = apply_header_mapping(df, mapping)

    # Validate required columns are present
    if not validate_required_columns(df):
        missing = set(ColName.mandatory_cols) - set(df.columns)
        raise ValueError(f"Cannot proceed: missing required columns {missing}")

    # Perform munging
    munged_df = munge(df)

    logger.info(f"Munging complete: {len(munged_df)} rows retained")

    return munged_df


def create_example_config(
    headers: List[str], output_path: str = "munge_config.json"
) -> str:
    """
    Create an example configuration file.

    Parameters
    ----------
    headers : List[str]
        List of column headers to create config for.
    output_path : str, optional
        Output path for configuration file.

    Returns
    -------
    str
        Path to created configuration file.
    """
    config = create_config_template(headers, interactive=False)

    with open(output_path, "w") as f:
        json.dump(config, f, indent=2)

    logger.info(f"Created example configuration: {output_path}")
    return output_path


def validate_config(config: Dict[str, Any]) -> bool:
    """
    Validate configuration dictionary.

    Parameters
    ----------
    config : Dict[str, Any]
        Configuration to validate.

    Returns
    -------
    bool
        True if configuration is valid.
    """
    required_keys = ["column_mapping"]

    for key in required_keys:
        if key not in config:
            logger.error(f"Missing required configuration key: {key}")
            return False

    # Validate column mapping
    mapping = config["column_mapping"]
    if not isinstance(mapping, dict):
        logger.error("column_mapping must be a dictionary")
        return False

    # Check if we can map required columns
    mapped_standards = set(mapping.values())
    required_standards = set(ColName.mandatory_cols)
    missing = required_standards - mapped_standards

    if missing:
        logger.error(f"Configuration missing mappings for required columns: {missing}")
        return False

    logger.info("Configuration validation passed")
    return True



"""
Core munging functions for GWAS summary statistics.

Adapted from smunger (https://github.com/Jianhua-Wang/smunger)
Original author: Jianhua Wang
License: MIT

This module provides the main data cleaning and standardization functions
for processing GWAS summary statistics.
"""

import logging
from typing import Optional

import numpy as np
import pandas as pd

from .constants import ColAllowNA, ColName, ColRange, ColType
from .validation import check_mandatory_cols, validate_and_clean_column

logger = logging.getLogger("Munging")


def munge(df: pd.DataFrame) -> pd.DataFrame:
    """
    Clean and standardize GWAS summary statistics.

    Adapted from smunger.munge() function.

    Parameters
    ----------
    df : pd.DataFrame
        Input DataFrame with GWAS summary statistics.

    Returns
    -------
    pd.DataFrame
        Cleaned and standardized DataFrame.

    Notes
    -----
    This function performs comprehensive data cleaning including:
    1. Removing columns with all NA values
    2. Cleaning and validating core columns (CHR, BP, alleles)
    3. Creating unique SNP identifiers
    4. Processing p-values, effect sizes, and other statistics
    5. Sorting by chromosome and position
    """
    logger.info("Starting munging process")
    original_rows = len(df)

    # Check mandatory columns
    check_mandatory_cols(df)

    # Make a copy to avoid modifying input
    outdf = df.copy()

    # Remove columns that are all NA
    outdf = _remove_all_na_columns(outdf)

    # Clean core columns
    outdf = _munge_chr(outdf)
    outdf = _munge_bp(outdf)
    outdf = _munge_alleles(outdf)

    # Create unique SNP identifiers
    outdf = make_SNPID_unique(outdf)

    # Process statistical columns
    outdf = _munge_pvalue(outdf)
    outdf = _munge_beta(outdf)
    outdf = _munge_se(outdf)

    # Process frequency columns if present
    if ColName.EAF in outdf.columns:
        outdf = _munge_eaf(outdf)

    # Sort by chromosome and position
    outdf = outdf.sort_values([ColName.CHR, ColName.BP])
    outdf.reset_index(drop=True, inplace=True)

    # Ensure correct column order and types
    outdf = _finalize_columns(outdf)

    final_rows = len(outdf)
    logger.info(
        f"Munging complete: {original_rows} -> {final_rows} rows ({original_rows - final_rows} removed)"
    )

    return outdf


def make_SNPID_unique(
    df: pd.DataFrame,
    remove_duplicates: bool = True,
    col_chr: str = ColName.CHR,
    col_bp: str = ColName.BP,
    col_ea: str = ColName.EA,
    col_nea: str = ColName.NEA,
    col_p: str = ColName.P,
) -> pd.DataFrame:
    """
    Generate unique SNP identifiers and optionally remove duplicates.

    Adapted from smunger.make_SNPID_unique() function.

    Parameters
    ----------
    df : pd.DataFrame
        Input DataFrame.
    remove_duplicates : bool, optional
        Whether to remove duplicated SNPs, keeping the one with smallest p-value.
    col_chr : str, optional
        Column name for chromosome.
    col_bp : str, optional
        Column name for base pair position.
    col_ea : str, optional
        Column name for effect allele.
    col_nea : str, optional
        Column name for non-effect allele.
    col_p : str, optional
        Column name for p-value.

    Returns
    -------
    pd.DataFrame
        DataFrame with unique SNPID column.
    """
    outdf = df.copy()

    # Sort alleles alphabetically to ensure consistent SNP IDs
    allele_df = outdf[[col_ea, col_nea]].apply(
        lambda row: sorted([str(row[col_ea]), str(row[col_nea])]),
        axis=1,
        result_type="expand",
    )
    allele_df.columns = ["allele1", "allele2"]

    # Create unique SNPID: chr-bp-allele1-allele2
    outdf[ColName.SNPID] = (
        outdf[col_chr].astype(str)
        + "-"
        + outdf[col_bp].astype(str)
        + "-"
        + allele_df["allele1"]
        + "-"
        + allele_df["allele2"]
    )

    # Move SNPID to first column
    cols = outdf.columns.tolist()
    if ColName.SNPID in cols:
        cols.insert(0, cols.pop(cols.index(ColName.SNPID)))
        outdf = outdf[cols]

    # Handle duplicates
    n_duplicated = outdf.duplicated(subset=[ColName.SNPID]).sum()

    if remove_duplicates and n_duplicated > 0:
        logger.info(f"Removing {n_duplicated} duplicate SNPs")
        if col_p in outdf.columns:
            # Sort by p-value to keep most significant
            outdf = outdf.sort_values(col_p)
        outdf = outdf.drop_duplicates(subset=[ColName.SNPID], keep="first")
        outdf = outdf.sort_values([col_chr, col_bp])
        outdf.reset_index(drop=True, inplace=True)
    elif n_duplicated > 0:
        logger.warning(f"Found {n_duplicated} duplicate SNPs, keeping all")
        # Add suffix to make unique
        dup_suffix = "-" + outdf.groupby(ColName.SNPID).cumcount().astype(str)
        dup_suffix = dup_suffix.str.replace("-0", "")
        outdf[ColName.SNPID] = outdf[ColName.SNPID] + dup_suffix

    logger.debug(f"Created unique SNPIDs: {len(outdf)} variants")

    return outdf


def _remove_all_na_columns(df: pd.DataFrame) -> pd.DataFrame:
    """Remove columns that are entirely NA."""
    outdf = df.copy()
    outdf = outdf.replace("", np.nan)  # Convert empty strings to NaN

    for col in outdf.columns:
        if outdf[col].isna().all():
            logger.debug(f"Removing all-NA column: {col}")
            outdf = outdf.drop(columns=[col])

    return outdf


def _munge_chr(df: pd.DataFrame) -> pd.DataFrame:
    """Clean and validate chromosome column."""
    return validate_and_clean_column(
        df=df,
        col_name=ColName.CHR,
        col_type=ColType.CHR,
        min_val=ColRange.CHR_MIN,
        max_val=ColRange.CHR_MAX,
        allow_na=ColAllowNA.CHR,
        transform_func=_transform_chr,
    )


def _transform_chr(series: pd.Series) -> pd.Series:
    """Transform chromosome values to standard format."""
    # Convert to string first
    result = series.astype(str)

    # Remove 'chr' prefix
    result = result.str.replace("chr", "", case=False)

    # Convert X to 23
    result = result.replace(["X", "x"], "23")

    # Convert to numeric
    result = pd.to_numeric(result, errors="coerce")

    return result


def _munge_bp(df: pd.DataFrame) -> pd.DataFrame:
    """Clean and validate base pair position column."""
    return validate_and_clean_column(
        df=df,
        col_name=ColName.BP,
        col_type=ColType.BP,
        min_val=ColRange.BP_MIN,
        max_val=ColRange.BP_MAX,
        allow_na=ColAllowNA.BP,
    )


def _munge_alleles(df: pd.DataFrame) -> pd.DataFrame:
    """Clean and validate allele columns."""
    outdf = df.copy()

    # Clean effect allele
    outdf = validate_and_clean_column(
        df=outdf,
        col_name=ColName.EA,
        col_type=ColType.EA,
        allow_na=ColAllowNA.EA,
        transform_func=_transform_allele,
    )

    # Clean non-effect allele
    outdf = validate_and_clean_column(
        df=outdf,
        col_name=ColName.NEA,
        col_type=ColType.NEA,
        allow_na=ColAllowNA.NEA,
        transform_func=_transform_allele,
    )

    # Remove rows where EA == NEA
    before_count = len(outdf)
    outdf = outdf[outdf[ColName.EA] != outdf[ColName.NEA]]
    after_count = len(outdf)

    if before_count > after_count:
        logger.debug(
            f"Removed {before_count - after_count} rows with identical alleles"
        )

    return outdf


def _transform_allele(series: pd.Series) -> pd.Series:
    """Transform allele values to standard format."""
    result = series.astype(str).str.upper()

    # Only keep valid DNA bases
    valid_pattern = r"^[ACGT]+$"
    mask = result.str.match(valid_pattern, na=False)
    result.loc[~mask] = np.nan

    return result


def _munge_pvalue(df: pd.DataFrame) -> pd.DataFrame:
    """Clean and validate p-value column."""
    return validate_and_clean_column(
        df=df,
        col_name=ColName.P,
        col_type=ColType.P,
        min_val=ColRange.P_MIN,
        max_val=ColRange.P_MAX,
        allow_na=ColAllowNA.P,
        exclude_min=True,  # P-value cannot be exactly 0
    )


def _munge_beta(df: pd.DataFrame) -> pd.DataFrame:
    """Clean and validate beta (effect size) column."""
    return validate_and_clean_column(
        df=df, col_name=ColName.BETA, col_type=ColType.BETA, allow_na=ColAllowNA.BETA
    )


def _munge_se(df: pd.DataFrame) -> pd.DataFrame:
    """Clean and validate standard error column."""
    return validate_and_clean_column(
        df=df,
        col_name=ColName.SE,
        col_type=ColType.SE,
        min_val=ColRange.SE_MIN,
        allow_na=ColAllowNA.SE,
        exclude_min=True,  # SE cannot be exactly 0
    )


def _munge_eaf(df: pd.DataFrame) -> pd.DataFrame:
    """Clean and validate effect allele frequency column."""
    return validate_and_clean_column(
        df=df,
        col_name=ColName.EAF,
        col_type=ColType.EAF,
        min_val=ColRange.EAF_MIN,
        max_val=ColRange.EAF_MAX,
        allow_na=ColAllowNA.EAF,
    )


def _munge_maf(df: pd.DataFrame) -> pd.DataFrame:
    """Clean and validate minor allele frequency column."""
    return validate_and_clean_column(
        df=df,
        col_name=ColName.MAF,
        col_type=ColType.MAF,
        min_val=ColRange.MAF_MIN,
        max_val=ColRange.MAF_MAX,
        allow_na=ColAllowNA.MAF,
    )


def _finalize_columns(df: pd.DataFrame) -> pd.DataFrame:
    """Ensure correct column order and add missing columns for credtools output."""
    outdf = df.copy()

    # Add missing output columns with None values
    for col in ColName.output_cols:
        if col not in outdf.columns:
            outdf[col] = None

    # Reorder columns to standard output order (CHR, BP, SNPID, EA, NEA, EAF, BETA, SE, P, RSID)
    outdf = outdf[ColName.output_cols]

    return outdf



"""Functions for processing summary statistics data."""

import gzip
import logging
from typing import Any, Optional

import numpy as np
import pandas as pd

from .constants import ColName, ColRange, ColType

logger = logging.getLogger("Sumstats")


def get_significant_snps(
    df: pd.DataFrame,
    pvalue_threshold: float = 5e-8,
    use_most_sig_if_no_sig: bool = True,
) -> pd.DataFrame:
    """
    Retrieve significant SNPs from the input DataFrame based on a p-value threshold.

    Parameters
    ----------
    df : pd.DataFrame
        The input summary statistics containing SNP information.
    pvalue_threshold : float, optional
        The p-value threshold for significance, by default 5e-8.
    use_most_sig_if_no_sig : bool, optional
        Whether to return the most significant SNP if no SNP meets the threshold, by default True.

    Returns
    -------
    pd.DataFrame
        A DataFrame containing significant SNPs, sorted by p-value in ascending order.

    Raises
    ------
    ValueError
        If no significant SNPs are found and `use_most_sig_if_no_sig` is False,
        or if the DataFrame is empty.
    KeyError
        If required columns are not present in the input DataFrame.

    Notes
    -----
    If no SNPs meet the significance threshold and `use_most_sig_if_no_sig` is True,
    the function returns the SNP with the smallest p-value.

    Examples
    --------
    >>> data = {
    ...     'SNPID': ['rs1', 'rs2', 'rs3'],
    ...     'P': [1e-9, 0.05, 1e-8]
    ... }
    >>> df = pd.DataFrame(data)
    >>> significant_snps = get_significant_snps(df, pvalue_threshold=5e-8)
    >>> print(significant_snps)
        SNPID         P
    0    rs1  1.000000e-09
    2    rs3  1.000000e-08
    """
    required_columns = {ColName.P, ColName.SNPID}
    missing_columns = required_columns - set(df.columns)
    if missing_columns:
        raise KeyError(
            f"The following required columns are missing from the DataFrame: {missing_columns}"
        )

    sig_df = df.loc[df[ColName.P] <= pvalue_threshold].copy()

    if sig_df.empty:
        if use_most_sig_if_no_sig:
            min_pvalue = df[ColName.P].min()
            sig_df = df.loc[df[ColName.P] == min_pvalue].copy()
            if sig_df.empty:
                raise ValueError("The DataFrame is empty. No SNPs available to select.")
            logging.debug(
                f"Using the most significant SNP: {sig_df.iloc[0][ColName.SNPID]}"
            )
            logging.debug(f"p-value: {sig_df.iloc[0][ColName.P]}")
        else:
            raise ValueError("No significant SNPs found.")
    else:
        sig_df.sort_values(by=ColName.P, inplace=True)
        sig_df.reset_index(drop=True, inplace=True)

    return sig_df


def make_SNPID_unique(
    sumstat: pd.DataFrame,
    remove_duplicates: bool = True,
    col_chr: str = ColName.CHR,
    col_bp: str = ColName.BP,
    col_ea: str = ColName.EA,
    col_nea: str = ColName.NEA,
    col_p: str = ColName.P,
) -> pd.DataFrame:
    """
    Generate unique SNP identifiers to facilitate the combination of multiple summary statistics datasets.

    Parameters
    ----------
    sumstat : pd.DataFrame
        The input summary statistics containing SNP information.
    remove_duplicates : bool, optional
        Whether to remove duplicated SNPs, keeping the one with the smallest p-value, by default True.
    col_chr : str, optional
        The column name for chromosome information, by default ColName.CHR.
    col_bp : str, optional
        The column name for base-pair position information, by default ColName.BP.
    col_ea : str, optional
        The column name for effect allele information, by default ColName.EA.
    col_nea : str, optional
        The column name for non-effect allele information, by default ColName.NEA.
    col_p : str, optional
        The column name for p-value information, by default ColName.P.

    Returns
    -------
    pd.DataFrame
        The summary statistics DataFrame with unique SNPIDs, suitable for merging with other datasets.

    Raises
    ------
    KeyError
        If required columns are missing from the input DataFrame.
    ValueError
        If the input DataFrame is empty or becomes empty after processing.

    Notes
    -----
    This function constructs a unique SNPID by concatenating chromosome, base-pair position,
    and sorted alleles (EA and NEA). This unique identifier allows for efficient merging of
    multiple summary statistics without the need for extensive duplicate comparisons.

    The unique SNPID format: "chr-bp-sortedEA-sortedNEA"

    If duplicates are found and `remove_duplicates` is False, a suffix "-N" is added to make
    identifiers unique, where N is the occurrence number.

    Examples
    --------
    >>> data = {
    ...     'CHR': ['1', '1', '2'],
    ...     'BP': [12345, 12345, 67890],
    ...     'EA': ['A', 'A', 'G'],
    ...     'NEA': ['G', 'G', 'A'],
    ...     'rsID': ['rs1', 'rs2', 'rs3'],
    ...     'P': [1e-5, 1e-6, 1e-7]
    ... }
    >>> df = pd.DataFrame(data)
    >>> unique_df = make_SNPID_unique(df, remove_duplicates=True)
    >>> print(unique_df)
        SNPID       CHR     BP EA NEA rsID         P
    0  1-12345-A-G    1  12345  A   G  rs2  1.000000e-06
    1  2-67890-A-G    2  67890  G   A  rs3  1.000000e-07
    """
    required_columns = {
        col_chr,
        col_bp,
        col_ea,
        col_nea,
    }
    missing_columns = required_columns - set(sumstat.columns)
    if missing_columns:
        raise KeyError(
            f"The following required columns are missing from the DataFrame: {missing_columns}"
        )

    if sumstat.empty:
        raise ValueError("The input DataFrame is empty.")

    df = sumstat.copy()

    # Sort alleles to ensure unique representation (EA <= NEA)
    allele_df = df[[col_ea, col_nea]].apply(
        lambda row: sorted([row[col_ea], row[col_nea]]), axis=1, result_type="expand"
    )
    allele_df.columns = [col_ea, col_nea]

    # Create unique SNPID
    df[ColName.SNPID] = (
        df[col_chr].astype(str)
        + "-"
        + df[col_bp].astype(str)
        + "-"
        + allele_df[col_ea]
        + "-"
        + allele_df[col_nea]
    )

    # move SNPID to the first column
    cols = df.columns.tolist()
    cols.insert(0, cols.pop(cols.index(ColName.SNPID)))
    df = df[cols]

    n_duplicated = df.duplicated(subset=[ColName.SNPID]).sum()

    if remove_duplicates and n_duplicated > 0:
        logger.debug(f"Number of duplicated SNPs: {n_duplicated}")
        if col_p in df.columns:
            # Sort by p-value to keep the SNP with the smallest p-value
            df.sort_values(by=col_p, inplace=True)
        df.drop_duplicates(subset=[ColName.SNPID], keep="first", inplace=True)
        # Sort DataFrame by chromosome and base-pair position
        df.sort_values(by=[col_chr, col_bp], inplace=True)
        df.reset_index(drop=True, inplace=True)
    elif n_duplicated > 0 and not remove_duplicates:
        logger.warning(
            """Duplicated SNPs detected. To remove duplicates, set `remove_duplicates=True`.
            Change the Unique SNP identifier to make it unique."""
        )
        # Change the Unique SNP identifier to make it unique. add a number to the end of the SNP identifier
        #  for example, 1-12345-A-G to 1-12345-A-G-1, 1-12345-A-G-2, etc. no alteration to the original SNP identifier
        dup_tail = "-" + df.groupby(ColName.SNPID).cumcount().astype(str)
        dup_tail = dup_tail.str.replace("-0", "")
        df[ColName.SNPID] = df[ColName.SNPID] + dup_tail

    logger.debug("Unique SNPIDs have been successfully created.")
    logger.debug(f"Total unique SNPs: {len(df)}")

    return df


def check_colnames(df: pd.DataFrame) -> pd.DataFrame:
    """
    Check column names in the DataFrame and fill missing columns with None.

    Parameters
    ----------
    df : pd.DataFrame
        Input DataFrame to check for column names.

    Returns
    -------
    pd.DataFrame
        DataFrame with all required columns, filling missing ones with None.

    Notes
    -----
    This function ensures that all required summary statistics columns are present
    in the DataFrame. Missing columns are added with None values.
    """
    outdf: pd.DataFrame = df.copy()
    for col in ColName.sumstat_cols:
        if col not in outdf.columns:
            outdf[col] = None
    return outdf[ColName.sumstat_cols]


def check_mandatory_cols(df: pd.DataFrame) -> None:
    """
    Check if the DataFrame contains all mandatory columns.

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame to check for mandatory columns.

    Returns
    -------
    None

    Raises
    ------
    ValueError
        If any mandatory columns are missing.

    Notes
    -----
    Mandatory columns are defined in ColName.mandatory_cols and typically include
    essential fields like chromosome, position, alleles, effect size, and p-value.
    """
    outdf = df.copy()
    missing_cols = set(ColName.mandatory_cols) - set(outdf.columns)
    if missing_cols:
        raise ValueError(f"Missing mandatory columns: {missing_cols}")
    return None


def rm_col_allna(df: pd.DataFrame) -> pd.DataFrame:
    """
    Remove columns from the DataFrame that are entirely NA.

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame from which to remove columns.

    Returns
    -------
    pd.DataFrame
        A DataFrame with columns that are entirely NA removed.

    Notes
    -----
    This function also converts empty strings to None before checking for
    all-NA columns. Columns that contain only missing values are dropped
    to reduce memory usage and improve processing efficiency.
    """
    outdf = df.copy()
    outdf = outdf.replace("", None)
    for col in outdf.columns:
        if outdf[col].isnull().all():
            logger.debug(f"Remove column {col} because it is all NA.")
            outdf.drop(col, axis=1, inplace=True)
    return outdf


def munge(df: pd.DataFrame) -> pd.DataFrame:
    """
    Munge the summary statistics DataFrame by performing a series of transformations.

    Parameters
    ----------
    df : pd.DataFrame
        The input DataFrame containing summary statistics.

    Returns
    -------
    pd.DataFrame
        The munged DataFrame with necessary transformations applied.

    Raises
    ------
    ValueError
        If any mandatory columns are missing.

    Notes
    -----
    This function performs comprehensive data cleaning and standardization:

    1. Validates mandatory columns are present
    2. Removes entirely missing columns
    3. Cleans chromosome and position data
    4. Validates and standardizes allele information
    5. Creates unique SNP identifiers
    6. Validates p-values, effect sizes, and standard errors
    7. Processes allele frequencies
    8. Handles rsID information if present

    The function applies strict quality control and may remove variants
    that don't meet validation criteria.
    """
    check_mandatory_cols(df)
    outdf = df.copy()
    outdf = rm_col_allna(outdf)
    outdf = munge_chr(outdf)
    outdf = munge_bp(outdf)
    outdf = munge_allele(outdf)
    outdf = make_SNPID_unique(outdf)
    outdf = munge_pvalue(outdf)
    outdf = outdf.sort_values(by=[ColName.CHR, ColName.BP])
    outdf = munge_beta(outdf)
    outdf = munge_se(outdf)
    outdf = munge_eaf(outdf)
    outdf[ColName.MAF] = outdf[ColName.EAF]
    outdf = munge_maf(outdf)
    if ColName.RSID in outdf.columns:
        outdf = munge_rsid(outdf)
    outdf = check_colnames(outdf)
    return outdf


def munge_rsid(df: pd.DataFrame) -> pd.DataFrame:
    """
    Munge rsID column.

    Parameters
    ----------
    df : pd.DataFrame
        Input DataFrame with rsID column.

    Returns
    -------
    pd.DataFrame
        DataFrame with munged rsID column.

    Notes
    -----
    This function converts the rsID column to the appropriate data type
    as defined in ColType.RSID.
    """
    outdf = df.copy()
    outdf[ColName.RSID] = outdf[ColName.RSID].astype(ColType.RSID)
    return outdf


def munge_chr(df: pd.DataFrame) -> pd.DataFrame:
    """
    Munge chromosome column.

    Parameters
    ----------
    df : pd.DataFrame
        Input DataFrame with chromosome column.

    Returns
    -------
    pd.DataFrame
        DataFrame with munged chromosome column.

    Notes
    -----
    This function:

    1. Removes rows with missing chromosome values
    2. Converts chromosome to string and removes 'chr' prefix
    3. Converts X chromosome to numeric value (23)
    4. Validates chromosome values are within acceptable range
    5. Converts to appropriate data type

    Invalid chromosome values are removed and logged.
    """
    pre_n = df.shape[0]
    outdf = df[df[ColName.CHR].notnull()].copy()
    outdf[ColName.CHR] = outdf[ColName.CHR].astype(str)
    outdf[ColName.CHR] = outdf[ColName.CHR].str.replace("chr", "")
    outdf[ColName.CHR] = outdf[ColName.CHR].replace(["X", "x"], 23)
    outdf[ColName.CHR] = pd.to_numeric(outdf[ColName.CHR], errors="coerce")
    outdf = outdf[outdf[ColName.CHR].notnull()]
    outdf = outdf[
        (outdf[ColName.CHR] >= ColRange.CHR_MIN)
        & (outdf[ColName.CHR] <= ColRange.CHR_MAX)
    ]
    after_n = outdf.shape[0]
    logger.debug(f"Remove {pre_n - after_n} rows because of invalid chromosome.")
    outdf[ColName.CHR] = outdf[ColName.CHR].astype(ColType.CHR)
    return outdf


def munge_bp(df: pd.DataFrame) -> pd.DataFrame:
    """
    Munge position column.

    Parameters
    ----------
    df : pd.DataFrame
        Input DataFrame with position column.

    Returns
    -------
    pd.DataFrame
        DataFrame with munged position column.

    Notes
    -----
    This function:

    1. Removes rows with missing position values
    2. Converts position to numeric type
    3. Validates positions are within acceptable range
    4. Converts to appropriate data type

    Invalid position values are removed and logged.
    """
    pre_n = df.shape[0]
    outdf = df[df[ColName.BP].notnull()].copy()
    outdf[ColName.BP] = pd.to_numeric(outdf[ColName.BP], errors="coerce")
    outdf = outdf[outdf[ColName.BP].notnull()]
    outdf = outdf[
        (outdf[ColName.BP] > ColRange.BP_MIN) & (outdf[ColName.BP] < ColRange.BP_MAX)
    ]
    after_n = outdf.shape[0]
    logger.debug(f"Remove {pre_n - after_n} rows because of invalid position.")
    outdf[ColName.BP] = outdf[ColName.BP].astype(ColType.BP)
    return outdf


def munge_allele(df: pd.DataFrame) -> pd.DataFrame:
    """
    Munge allele columns.

    Parameters
    ----------
    df : pd.DataFrame
        Input DataFrame with allele columns.

    Returns
    -------
    pd.DataFrame
        DataFrame with munged allele columns.

    Notes
    -----
    This function:

    1. Removes rows with missing allele values
    2. Converts alleles to uppercase
    3. Validates alleles contain only valid DNA bases (A, C, G, T)
    4. Removes variants where effect allele equals non-effect allele

    Invalid alleles and monomorphic variants are removed and logged.
    """
    outdf = df.copy()
    for col in [ColName.EA, ColName.NEA]:
        pre_n = outdf.shape[0]
        outdf = outdf[outdf[col].notnull()]
        outdf[col] = outdf[col].astype(str).str.upper()
        outdf = outdf[outdf[col].str.match(r"^[ACGT]+$")]
        after_n = outdf.shape[0]
        logger.debug(f"Remove {pre_n - after_n} rows because of invalid {col}.")
    outdf = outdf[outdf[ColName.EA] != outdf[ColName.NEA]]
    return outdf


def munge_pvalue(df: pd.DataFrame) -> pd.DataFrame:
    """
    Munge p-value column.

    Parameters
    ----------
    df : pd.DataFrame
        Input DataFrame with p-value column.

    Returns
    -------
    pd.DataFrame
        DataFrame with munged p-value column.

    Notes
    -----
    This function:

    1. Converts p-values to numeric type
    2. Removes rows with missing p-values
    3. Validates p-values are within acceptable range (0, 1)
    4. Converts to appropriate data type

    Invalid p-values are removed and logged.
    """
    outdf = df.copy()
    pre_n = outdf.shape[0]
    outdf[ColName.P] = pd.to_numeric(outdf[ColName.P], errors="coerce")
    outdf = outdf[outdf[ColName.P].notnull()]
    outdf = outdf[
        (outdf[ColName.P] > ColRange.P_MIN) & (outdf[ColName.P] < ColRange.P_MAX)
    ]
    after_n = outdf.shape[0]
    logger.debug(f"Remove {pre_n - after_n} rows because of invalid p-value.")
    outdf[ColName.P] = outdf[ColName.P].astype(ColType.P)
    return outdf


def munge_beta(df: pd.DataFrame) -> pd.DataFrame:
    """
    Munge beta column.

    Parameters
    ----------
    df : pd.DataFrame
        Input DataFrame with beta column.

    Returns
    -------
    pd.DataFrame
        DataFrame with munged beta column.

    Notes
    -----
    This function:

    1. Converts beta values to numeric type
    2. Removes rows with missing beta values
    3. Converts to appropriate data type

    Invalid beta values are removed and logged.
    """
    pre_n = df.shape[0]
    outdf = df.copy()
    outdf[ColName.BETA] = pd.to_numeric(outdf[ColName.BETA], errors="coerce")
    outdf = outdf[outdf[ColName.BETA].notnull()]
    after_n = outdf.shape[0]
    logger.debug(f"Remove {pre_n - after_n} rows because of invalid beta.")
    outdf[ColName.BETA] = outdf[ColName.BETA].astype(ColType.BETA)
    return outdf


def munge_se(df: pd.DataFrame) -> pd.DataFrame:
    """
    Munge standard error column.

    Parameters
    ----------
    df : pd.DataFrame
        Input DataFrame with standard error column.

    Returns
    -------
    pd.DataFrame
        DataFrame with munged standard error column.

    Notes
    -----
    This function:

    1. Converts standard error values to numeric type
    2. Removes rows with missing standard error values
    3. Validates standard errors are positive
    4. Converts to appropriate data type

    Invalid standard error values are removed and logged.
    """
    pre_n = df.shape[0]
    outdf = df.copy()
    outdf[ColName.SE] = pd.to_numeric(outdf[ColName.SE], errors="coerce")
    outdf = outdf[outdf[ColName.SE].notnull()]
    outdf = outdf[outdf[ColName.SE] > ColRange.SE_MIN]
    after_n = outdf.shape[0]
    logger.debug(f"Remove {pre_n - after_n} rows because of invalid standard error.")
    outdf[ColName.SE] = outdf[ColName.SE].astype(ColType.SE)
    return outdf


def munge_eaf(df: pd.DataFrame) -> pd.DataFrame:
    """
    Munge effect allele frequency column.

    Parameters
    ----------
    df : pd.DataFrame
        Input DataFrame with effect allele frequency column.

    Returns
    -------
    pd.DataFrame
        DataFrame with munged effect allele frequency column.

    Notes
    -----
    This function:

    1. Converts EAF values to numeric type
    2. Removes rows with missing EAF values
    3. Validates EAF values are within range [0, 1]
    4. Converts to appropriate data type

    Invalid EAF values are removed and logged.
    """
    pre_n = df.shape[0]
    outdf = df.copy()
    outdf[ColName.EAF] = pd.to_numeric(outdf[ColName.EAF], errors="coerce")
    outdf = outdf[outdf[ColName.EAF].notnull()]
    outdf = outdf[
        (outdf[ColName.EAF] >= ColRange.EAF_MIN)
        & (outdf[ColName.EAF] <= ColRange.EAF_MAX)
    ]
    after_n = outdf.shape[0]
    logger.debug(
        f"Remove {pre_n - after_n} rows because of invalid effect allele frequency."
    )
    outdf[ColName.EAF] = outdf[ColName.EAF].astype(ColType.EAF)
    return outdf


def munge_maf(df: pd.DataFrame) -> pd.DataFrame:
    """
    Munge minor allele frequency column.

    Parameters
    ----------
    df : pd.DataFrame
        Input DataFrame with minor allele frequency column.

    Returns
    -------
    pd.DataFrame
        DataFrame with munged minor allele frequency column.

    Notes
    -----
    This function:

    1. Converts MAF values to numeric type
    2. Removes rows with missing MAF values
    3. Converts frequencies > 0.5 to 1 - frequency (to ensure minor allele)
    4. Validates MAF values are within acceptable range
    5. Converts to appropriate data type

    Invalid MAF values are removed and logged.
    """
    pre_n = df.shape[0]
    outdf = df.copy()
    outdf[ColName.MAF] = pd.to_numeric(outdf[ColName.MAF], errors="coerce")
    outdf = outdf[outdf[ColName.MAF].notnull()]
    outdf[ColName.MAF] = outdf[ColName.MAF].apply(lambda x: 1 - x if x > 0.5 else x)
    outdf = outdf[
        (outdf[ColName.MAF] >= ColRange.MAF_MIN)
        & (outdf[ColName.MAF] <= ColRange.MAF_MAX)
    ]
    after_n = outdf.shape[0]
    logger.debug(
        f"Remove {pre_n - after_n} rows because of invalid minor allele frequency."
    )
    outdf[ColName.MAF] = outdf[ColName.MAF].astype(ColType.MAF)
    return outdf


def sort_alleles(df: pd.DataFrame) -> pd.DataFrame:
    """
    Sort EA and NEA in alphabetical order. Change the sign of beta if EA is not sorted as the first allele.

    Parameters
    ----------
    df : pd.DataFrame
        Input DataFrame with allele columns.

    Returns
    -------
    pd.DataFrame
        DataFrame with sorted allele columns.

    Notes
    -----
    This function ensures consistent allele ordering by:

    1. Sorting effect allele (EA) and non-effect allele (NEA) alphabetically
    2. Flipping the sign of beta if alleles were swapped
    3. Adjusting effect allele frequency (EAF) if alleles were swapped (EAF = 1 - EAF)

    This standardization is important for:
    - Consistent merging across datasets
    - Meta-analysis compatibility
    - LD matrix alignment
    """
    outdf = df.copy()
    outdf[["sorted_a1", "sorted_a2"]] = np.sort(
        outdf[[ColName.EA, ColName.NEA]], axis=1
    )
    outdf[ColName.BETA] = np.where(
        outdf[ColName.EA] == outdf["sorted_a1"],
        outdf[ColName.BETA],
        -outdf[ColName.BETA],
    )
    if ColName.EAF in outdf.columns:
        outdf[ColName.EAF] = np.where(
            outdf[ColName.EA] == outdf["sorted_a1"],
            outdf[ColName.EAF],
            1 - outdf[ColName.EAF],
        )
    outdf[ColName.EA] = outdf["sorted_a1"]
    outdf[ColName.NEA] = outdf["sorted_a2"]
    outdf.drop(columns=["sorted_a1", "sorted_a2"], inplace=True)
    return outdf


def load_sumstats(
    filename: str,
    if_sort_alleles: bool = True,
    sep: Optional[str] = None,
    nrows: Optional[int] = None,
    skiprows: int = 0,
    comment: Optional[str] = None,
    gzipped: Optional[bool] = None,
) -> pd.DataFrame:
    """
    Load summary statistics from a file.

    Parameters
    ----------
    filename : str
        The path to the file containing the summary statistics.
        The header must contain the column names: CHR, BP, EA, NEA, EAF, BETA, SE, P.
    if_sort_alleles : bool, optional
        Whether to sort alleles in alphabetical order, by default True.
    sep : Optional[str], optional
        The delimiter to use. If None, the delimiter is inferred from the file, by default None.
    nrows : Optional[int], optional
        Number of rows to read. If None, all rows are read, by default None.
    skiprows : int, optional
        Number of lines to skip at the start of the file, by default 0.
    comment : Optional[str], optional
        Character to split comments in the file, by default None.
    gzipped : Optional[bool], optional
        Whether the file is gzipped. If None, it is inferred from the file extension, by default None.

    Returns
    -------
    pd.DataFrame
        A DataFrame containing the loaded summary statistics.

    Notes
    -----
    The function performs the following operations:

    1. Auto-detects file compression (gzip) from file extension
    2. Auto-detects delimiter (tab, comma, or space) from file content
    3. Loads the data using pandas.read_csv
    4. Applies comprehensive data munging and quality control
    5. Optionally sorts alleles for consistency

    The function infers the delimiter if not provided and handles gzipped files automatically.
    Comprehensive quality control is applied including validation of chromosomes, positions,
    alleles, p-values, effect sizes, and frequencies.

    Examples
    --------
    >>> # Load summary statistics with automatic format detection
    >>> sumstats = load_sumstats('gwas_results.txt.gz')
    >>> print(f"Loaded {len(sumstats)} variants")
    Loaded 1000000 variants

    >>> # Load with specific parameters
    >>> sumstats = load_sumstats('gwas_results.csv', sep=',', nrows=10000)
    >>> print(sumstats.columns.tolist())
    ['SNPID', 'CHR', 'BP', 'EA', 'NEA', 'EAF', 'BETA', 'SE', 'P', 'MAF', 'RSID']
    """
    # determine whether the file is gzipped
    if gzipped is None:
        gzipped = filename.endswith("gz")

    # read the first line of the file to determine the separator
    if sep is None:
        if gzipped:
            f = gzip.open(filename, "rt")

        else:
            f = open(filename, "rt")
        if skiprows > 0:
            for _ in range(skiprows):
                f.readline()
        line = f.readline()
        f.close()
        if "\t" in line:
            sep = "\t"
        elif "," in line:
            sep = ","
        else:
            sep = " "
    logger.debug(f"File {filename} is gzipped: {gzipped}")
    logger.debug(f"Separator is {sep}")
    logger.debug(f"loading data from {filename}")
    # determine the separator, automatically if not specified
    sumstats = pd.read_csv(
        filename,
        sep=sep,
        nrows=nrows,
        skiprows=skiprows,
        comment=comment,
        compression="gzip" if gzipped else None,
    )
    sumstats = munge(sumstats)
    logger.info(f"Loaded {len(sumstats)} rows sumstats from {filename}")
    if if_sort_alleles:
        sumstats = sort_alleles(sumstats)
    return sumstats



"""Wrapper of SuSiE fine-mapping method."""

import json
import logging
from typing import List

import pandas as pd

from credtools.constants import ColName, Method
from credtools.credibleset import CredibleSet
from credtools.locus import Locus, intersect_sumstat_ld
from credtools.wrappers.susie_rss import susie_rss

logger = logging.getLogger("SuSiE")


def run_susie(
    locus: Locus,
    max_causal: int = 1,
    coverage: float = 0.95,
    max_iter: int = 100,
    estimate_residual_variance: bool = False,
    min_abs_corr: float = 0.1,
    convergence_tol: float = 1e-3,
) -> CredibleSet:
    """
    Run SuSiE (Sum of Single Effects) fine-mapping with shotgun stochastic search.

    SuSiE is a Bayesian variable selection method that fits a regression model
    with multiple single-effect components. Each component captures one causal
    signal, allowing the method to identify multiple independent associations
    while accounting for linkage disequilibrium.

    Parameters
    ----------
    locus : Locus
        Locus object containing summary statistics and LD matrix data.
        Must have matched summary statistics and LD matrix.
    max_causal : int, optional
        Maximum number of causal variants (L parameter), by default 1.
        This determines the number of single-effect components in the model.
        Higher values allow detection of more independent signals.
    coverage : float, optional
        Coverage probability for credible sets, by default 0.95.
        This determines the cumulative posterior probability mass
        included in each credible set.
    max_iter : int, optional
        Maximum number of iterations for the IBSS algorithm, by default 100.
        More iterations may improve convergence but increase runtime.
    estimate_residual_variance : bool, optional
        Whether to estimate residual variance from data, by default False.
        If False, residual variance is set to 1 (appropriate for z-scores).
    min_abs_corr : float, optional
        Minimum absolute correlation threshold for credible set purity, by default 0.1.
        Credible sets with pairwise correlations below this threshold may be filtered.
    convergence_tol : float, optional
        Convergence tolerance for the ELBO (Evidence Lower BOund), by default 1e-3.
        Algorithm stops when ELBO change falls below this threshold.

    Returns
    -------
    CredibleSet
        Credible set object containing:
        - Posterior inclusion probabilities for all variants
        - Credible sets for each detected signal
        - Lead SNPs (highest PIP in each credible set)
        - Coverage probabilities and purity measures
        - Algorithm parameters and convergence information

    Warnings
    --------
    If the summary statistics and LD matrix are not matched, they will be
    automatically intersected and reordered with a warning message.

    Notes
    -----
    SuSiE implements the "Sum of Single Effects" model:

    y = Σ(l=1 to L) X * b_l + ε

    where:
    - y is the phenotype vector
    - X is the genotype matrix
    - b_l is the l-th single-effect vector (sparse, at most one non-zero element)
    - ε is the residual error
    - L is the maximum number of causal variants

    Key features of the algorithm:

    1. **Iterative Bayesian Stepwise Selection (IBSS)**: Iteratively updates
       each single-effect component while holding others fixed

    2. **Automatic multiplicity control**: Each component models exactly one
       causal variant, naturally controlling for multiple testing

    3. **Credible sets with coverage guarantees**: Provides sets of variants
       with specified coverage probability for each signal

    4. **LD-aware inference**: Properly accounts for correlation structure
       through the LD matrix

    The algorithm workflow:
    1. Initialize L single-effect regressions with uniform priors
    2. Iteratively update each effect using variational Bayes
    3. Monitor convergence using Evidence Lower Bound (ELBO)
    4. Construct credible sets based on posterior inclusion probabilities
    5. Filter credible sets based on purity criteria

    Advantages:
    - Principled uncertainty quantification through credible sets
    - Natural handling of multiple causal variants
    - Robust to LD structure and population stratification
    - Computationally efficient for genome-wide analysis

    Reference:
    Wang, G. et al. A simple new approach to variable selection in regression,
    with application to genetic fine mapping. J. R. Stat. Soc. B 82, 1273-1300 (2020).

    Examples
    --------
    >>> # Basic SuSiE analysis with single causal variant
    >>> credible_set = run_susie(locus)
    >>> print(f"Found {credible_set.n_cs} credible sets")
    >>> print(f"Top PIP: {credible_set.pips.max():.4f}")
    Found 1 credible sets
    Top PIP: 0.8234

    >>> # SuSiE with multiple causal variants and strict convergence
    >>> credible_set = run_susie(
    ...     locus,
    ...     max_causal=5,
    ...     coverage=0.99,
    ...     max_iter=500,
    ...     convergence_tol=1e-6
    ... )
    >>> print(f"Detected {credible_set.n_cs} independent signals")
    >>> print(f"Credible set sizes: {credible_set.cs_sizes}")
    Detected 3 independent signals
    Credible set sizes: [12, 8, 5]

    >>> # Access posterior inclusion probabilities
    >>> top_variants = credible_set.pips.nlargest(10)
    >>> print("Top 10 variants by PIP:")
    >>> print(top_variants)
    Top 10 variants by PIP:
    rs123456    0.8234
    rs789012    0.7456
    rs345678    0.6789
    ...

    >>> # Examine credible sets
    >>> for i, snps in enumerate(credible_set.snps):
    ...     print(f"Credible set {i+1}: {len(snps)} variants")
    ...     print(f"Lead SNP: {credible_set.lead_snps[i]}")
    Credible set 1: 12 variants
    Lead SNP: rs123456
    Credible set 2: 8 variants
    Lead SNP: rs789012
    """
    if not locus.is_matched:
        logger.warning(
            "The sumstat and LD are not matched, will match them in same order."
        )
        locus = intersect_sumstat_ld(locus)
    logger.info(f"Running SuSiE on {locus}")
    parameters = {
        "max_causal": max_causal,
        "coverage": coverage,
        "max_iter": max_iter,
        "estimate_residual_variance": estimate_residual_variance,
        "min_abs_corr": min_abs_corr,
        "convergence_tol": convergence_tol,
    }
    logger.info(f"Parameters: {json.dumps(parameters, indent=4)}")
    s = susie_rss(
        bhat=locus.sumstats[ColName.BETA].to_numpy(),
        shat=locus.sumstats[ColName.SE].to_numpy(),
        n=locus.sample_size,
        R=locus.ld.r,
        L=max_causal,
        coverage=coverage,
        max_iter=max_iter,
        estimate_residual_variance=estimate_residual_variance,
        min_abs_corr=min_abs_corr,
        tol=convergence_tol,
    )
    pip = s["pip"]
    if s["sets"]["cs"]:
        cs_idx = list(s["sets"]["cs"].values())
        n_cs = len(cs_idx)
        cs_sizes = [len(idx) for idx in cs_idx]
        cred_snps = [locus.sumstats[ColName.SNPID].iloc[idx].tolist() for idx in cs_idx]
    else:
        n_cs = 0
        cs_sizes = []
        cred_snps = []
    pips = pd.Series(data=pip, index=locus.sumstats[ColName.SNPID].tolist())
    lead_snps = [str(pips[pips.index.isin(cred_snps[i])].idxmax()) for i in range(n_cs)]
    logger.info(f"Finished SuSiE on {locus}")
    logger.info(f"N of credible set: {n_cs}")
    logger.info(f"Credible set size: {cs_sizes}")
    return CredibleSet(
        tool=Method.SUSIE,
        n_cs=n_cs,
        coverage=coverage,
        lead_snps=lead_snps,
        snps=cred_snps,
        cs_sizes=cs_sizes,
        pips=pips,
        parameters=parameters,
    )



"""Wrapper for MultiSuSiE."""

import functools
import logging
import random
import sys

import numba
import numpy as np
from scipy.optimize import minimize_scalar
from tqdm import tqdm

logger = logging.getLogger("MULTISUSIE")


class S:
    """Class to hold MultiSuSiE results."""

    def __init__(
        self,
        pop_sizes,
        L,
        XTX_list,
        scaled_prior_variance,
        residual_variance,
        varY,
        prior_weights,
        float_type=np.float32,
    ):

        # code from init_setup
        num_pop = len(pop_sizes)
        p = XTX_list[0].shape[0]
        self.alpha = np.zeros((L, p), dtype=float_type) + 1.0 / p
        self.mu = np.zeros((num_pop, L, p), dtype=float_type)
        self.mu2 = np.zeros((num_pop, num_pop, L, p), dtype=float_type)
        self.Xr_list = [np.zeros(XTX.shape[0], dtype=float_type) for XTX in XTX_list]
        self.sigma2 = residual_variance.astype(float_type)
        self.pi = prior_weights.astype(float_type)
        self.n = np.array(pop_sizes)
        self.L = L

        # code from init_finalize
        self.V = scaled_prior_variance * varY + np.zeros((num_pop, L), dtype=float_type)
        self.V = self.V.astype(float_type)
        assert np.all(self.V >= 0)
        self.ER2 = np.zeros(num_pop, dtype=float_type) + np.nan
        self.KL = np.zeros(L, dtype=float_type) + np.nan
        self.lbf = np.zeros(L, dtype=float_type) + np.nan

        self.converged = False


class SER_RESULTS:
    """Class to hold Single Effect Regression results."""

    def __init__(self, alpha, mu, mu2, lbf, lbf_model, V):
        self.alpha = alpha
        self.mu = mu
        self.mu2 = mu2
        self.lbf = lbf
        self.V = V
        self.lbf_model = lbf_model


def multisusie_rss(
    R_list,
    population_sizes,
    b_list=None,
    s_list=None,
    z_list=None,
    varY_list=None,
    rho=0.75,
    L=10,
    scaled_prior_variance=0.2,
    prior_weights=None,
    standardize=False,
    pop_spec_standardization=True,
    estimate_residual_variance=True,
    estimate_prior_variance=True,
    estimate_prior_method="early_EM",
    pop_spec_effect_priors=True,
    iter_before_zeroing_effects=5,
    prior_tol=1e-9,
    max_iter=100,
    tol=1e-3,
    verbose=False,
    coverage=0.95,
    min_abs_corr=0.0,
    float_type=np.float32,
    low_memory_mode=False,
    recover_R=False,
    single_population_mac_thresh=20,
    mac_list=None,
    multi_population_maf_thresh=0,
    maf_list=None,
):
    r"""
    Top-level function for running MultiSuSiE.

    This function takes takes standard GWAS summary statistics, converts
    them to sufficient statistics, and runs MultiSuSiE on them.

    Parameters
    ----------
    multisusie_rss accepts two combinations of input parameters:
    1. b_list, s_list, R_list, varY_list, rho, population_sizes
        This is the preferred input format and is used in the MultiSuSiE paper.
    2. z_list, R_list, rho, population_sizes
        Here, we assume that both the genotypes and phenotypes have been
        standardized to have variance 1 within each population.  It's extremely
        important to censor low MAF variants in the population where they are rare.
        This is done by setting values in z_list for these variants to 0, and
        columns and rows corresponding to these variants to 0. multisusie_rss
        will not do this for you (but will with the first input combination)
        because it doesn't have access to the information that would allow us to
        approximate the minor allele frequency or minor allele count.
        It's also kind of weird to standardize the genotypes within each
        population beacuse variants that are rare in one population and common
        in other other will be assigned huge effect sizes.

    R_list: length K list of PxP numpy arrays representing the LD correlation
        matrices for each population. Each array should correspond
        to the same set of P variants, in the same order. Variants that should
        not be included in a given population due to low MAF can be assigned a
        value of np.nan.
    population_sizes: list of integers representing the number of samples in
        the GWAS for each population
    b_list: length K list of length P numpy arrays containing effect sizes,
        one for each population. Each array should correspond
        to the same set of P variants, in the same order. Variants that should
        not be included in a given population due to low MAF can be assigned a
        value of np.nan. Provide exactly one of (b_list and s_list) and z_list.
    s_list: length K list of length P numpy arrays containing effect size
        standard errors, one for each population. Each array should correspond
        to the same set of P variants, in the same order. Variants that should
        not be included in a given population due to low MAF can be assigned a
        value of np.nan.Provide exactly one of (b_list and s_list) and z_list.
    z_list: length K list of length P numpy arrays containing Z-scores, one for
        each population. Each array should correspond to the same set of P
        variants, in the same order. Variants that should not be included in a
        given population due to low MAF can be assigned a value of np.nan.
        Provide exactly one of (b_list and s_list) and z_list.
    varY_list: length K list representing the sample variance of the outcome
        in each population
    rho: PxP numpy array representing the effect size correlation matrix. In
        the manuscript, we show that this parameter has little impact on the
        estimated PIPs in practice.
    L: integer representing the maximum number of causal variants
    scaled_prior_variance: float representing the effect size prior variance,
        scaled by the residual variance. It's fine to set this to a number
        larger than what you expect the squared effect size to be (like the
        default value of 0.2) as long as estimate_prior_variance is set to True
        and estimate_prior_method is not set to None.
    prior_weights: numpy P-array of floats representing the prior probability
        of causality for each variant. Give None to use a uniform prior
    standardize: boolean, whether to adjust summmary statistics to be as if
        genotypes were standardized to have mean 0 and variance 1
    pop_spec_standardization: boolean, if standardize is True, whether to
        adjust summary statistics to be as if genotypes were standardized
        separately for each population, or pooled and then standardized
    estimate_residual_variance: boolean, whether to estimate the residual variance, $\\sigma^2_k$ in the manuscript
    estimate_prior_variance: boolean, whether to estimate the prior variance,
        $A^{(l)}$ in the manuscript
    estimate_prior_method: string, method to estimate the prior variance. Recommended
        values are 'early_EM' or None
    pop_spec_effect_priors: boolean, whether to estimate separate prior
        variance parameters for each population
    iter_before_zeroing_effects: integer, number of iterations to run before
        zeroing out component-population pairs (or components if
        pop_spec_effect_priors is False) that have a lower likelihood than a
        null model
    prior_tol: float which places a filter on the minimum prior variance
        for a component to be included when estimating PIPs
    max_iter: integer, maximum number of iterations to run
    tol: float, after iter_before_zeroing_effects iterations, results
        are returned if the ELBO increases by less than tol in an ieration
    verbose: boolean which indicates if an progress bar should be displayed
    coverage: float representing the minimum coverage of credible sets
    min_abs_corr: float representing the minimum absolute correlation between
        any pair of variants in a credible set. For each pair of variants,
        the max is taken across ancestries. In the case where min_abs_corr = 0,
        low_memory_mode = True, and recover_R = False, the purity of credible
        sets will not be calculated.
    float_type: numpy float type used. Set to np.float32 to minimize memory
        consumption
    low_memory_mode: boolean, if True, the input R_list will be modified in place.
        Decreases memory consumption by a factor of two. BUT, THE INPUT R_list
        WILL BE OVERWRITTEN. If you need R_list, set low_memory_mode to False
    recover_R: boolean, if True, the R matrices will be recovered from XTX,
        BUT variants with MAC/MAF estimate less than mac_filter/maf_filter will
        be censored.
    single_population_mac_thresh: float, variants with minor allele count less
        than this value in a single population will be censored in that population.
        The variant will not be censored in the other populations.
        If mac_list is not None, mac_list is used for the filtering.
        If mac_list is None, but maf_list is not, mac is caclulated using
        maf_list and population_sizes. Otherwise, if mac_list is None,
        maf_list is None, and b_list and s_list are provided as input, then
        minor allele counts will be estimated under an assumption of
        Hardy-Weinberg equilibrium.
    mac_list: length K list of floats representing the minor allele count for
        each variant in each population.
    multi_population_maf_thresh: float, variants with minor allele frequency
        less than this value in ALL poopulations will be censored. Note that
        it's probably faster computationally to do this before calling this
        function. If maf_list is not None, maf_list is used for the filtering.
        If maf_list is None, but mac_list is not, maf is caclulated using
        mac_list and population_sizes. Otherwise, if maf_list is None,
        mac_list is None, and b_list and s_list are provided as input, then
        minor allele frequencies will be estimated under an assumption of
        Hardy-Weinberg equilibrium.
    maf_list: length K list of floats representing the minor allele frequency for
        each variant in each population.

    Returns
    -------
    an object containing results with the following attributes:
        alpha: L x P numpy array of single-effect regression posterior
            inclusion probabilities
        mu: K x L x P numpy array of single-effect regression effect size
            posterior means, conditional on each variant being the causal variant
        mu2: K x K x L x P numpy array of single-effect regression effect size
            posterior seconds moments, conditional on each variant being the
            causal variant
        sigma2: length-K numpy array of residual variance estimates
        pi: length-P numpy array of prior inclusion probabilities
        n: length-K numpy array of sample sizes
        L: integer representing the maximum number of causal variants
        V: K x L numpy array of effect size prior variance estimates
        ER2: length-K numpy array of expected squared residuals
        KL: L x 1 numpy array of Kullback-Leibler divergences for each single
            effect regression
        lbf: L x 1 numpy array of log Bayes factors for each single effect
            regression
        converged: boolean indicating whether the algorithm converged

    """
    # provide exactly one of (b_list and s_list) and z_list
    if (z_list is not None) & ((b_list is not None) or (s_list is not None)):
        raise ValueError(
            "provide either (b_list and s_list) or z_list, but not both. "
            + "See the parameters section of help(multisusie_rss) for more information"
        )
    if (z_list is None) & (b_list is None) & (s_list is None):
        raise ValueError("provide either (b_list and s_list) or z_list, but not both")
    if (b_list is not None) & ((s_list is None) | (varY_list is None)):
        raise ValueError(
            "if b_list is provided, s_list and varY_list must also be provided"
        )
    if population_sizes is None:
        raise ValueError("population_sizes must be provided")

    # check input list lengths
    K = len(R_list)
    if z_list is None:
        assert len(s_list) == K  # type: ignore
        assert len(b_list) == K  # type: ignore
    else:
        assert len(z_list) == K
    if varY_list is not None:
        assert len(varY_list) == K
    assert len(population_sizes) == K
    assert np.isscalar(rho) | ((rho.shape[0] == K) & (rho.shape[1] == K))  # type: ignore

    # make copies of R if low_memory_mode=False to avoid modifying the input data
    if low_memory_mode:
        R_list_copy = R_list
        if z_list is None:
            b_list_copy = b_list
            s_list_copy = s_list
        else:
            z_list_copy = z_list
        logger.info(
            "low memory mode is on. THE INPUT R MATRICES HAVE BEEN "
            + "TRANSFORMED INTO XTX AND CENSORED BASED ON MISSINGNESS. "
            + "THE INPUT R MATRICES HAVE BEEN CHANGED. The datatypes of "
            + "b_list, s_list, and R_list have been mutated in place "
            + "to match float_type"
        )
    else:
        R_list_copy = [np.copy(R) for R in R_list]
        if z_list is None:
            b_list_copy = [np.copy(b) for b in b_list]  # type: ignore
            s_list_copy = [np.copy(s) for s in s_list]  # type: ignore
        else:
            z_list_copy = [np.copy(z) for z in z_list]

    # convert everything to the requested float type if necessary
    for i in range(K):
        if z_list is None:
            if b_list_copy[i].dtype != float_type:  # type: ignore
                b_list_copy[i] = b_list_copy[i].astype(float_type, copy=False)  # type: ignore
            if s_list_copy[i].dtype != float_type:  # type: ignore
                s_list_copy[i] = s_list_copy[i].astype(float_type, copy=False)  # type: ignore
        else:
            if z_list_copy[i].dtype != float_type:
                z_list_copy[i] = z_list_copy[i].astype(float_type, copy=False)
        if R_list_copy[i].dtype != float_type:
            R_list_copy[i] = R_list_copy[i].astype(float_type, copy=False)
        if (not np.isscalar(rho)) and (rho.dtype != float_type):  # type: ignore
            rho = rho.astype(float_type, copy=True)  # type: ignore

    population_sizes = np.array(population_sizes, dtype=float_type)
    if varY_list is None:
        varY_list = np.ones(K, dtype=float_type)
    else:
        varY_list = np.array(varY_list, dtype=float_type)
    YTY_list = []
    for i in range(K):
        YTY = (varY_list[i] * (population_sizes[i] - 1)).astype(float_type)
        YTY_list.append(YTY)

    # convert GWAS summary statistics to MultiSuSiE sufficient statistics
    XTX_list = []
    XTY_list = []
    if z_list is None:
        for i in range(K):
            # this function mutates R_list_copy[i]
            XTX, XTY = recover_XTX_and_XTY(
                b=b_list_copy[i], s=s_list_copy[i], R=R_list_copy[i], YTY=YTY_list[i], n=population_sizes[i]  # type: ignore
            )
            XTX_list.append(XTX)
            XTY_list.append(XTY)
    else:
        for i in range(K):
            # this function mutates R_list_copy[i]
            XTX, XTY = recover_XTX_and_XTY_from_Z(
                z=z_list_copy[i],
                R=R_list_copy[i],
                n=population_sizes[i],
                float_type=float_type,
            )
            XTX_list.append(XTX)
            XTY_list.append(XTY)

    # do minor allele count filtering for each population
    if single_population_mac_thresh > 0:
        for i in range(K):
            n = population_sizes[i]
            if mac_list is not None:
                mac = np.minimum(mac_list[i], 2 * n - mac_list[i])
            elif maf_list is not None:
                mac = np.minimum(2 * n * maf_list[i], 2 * n - 2 * n * maf_list[i])
            elif b_list is not None:
                var_x = np.minimum(
                    np.diag(XTX_list[i]) / (n - 1), 0.5
                )  # this can be > .5 due to HWE violations?
                maf = 1 / 2 - np.sqrt(1 - 2 * var_x) / 2
                mac = np.minimum(2 * n * maf, 2 * n - 2 * n * maf)
            else:
                logger.info(
                    "Skipping MAC filtering because mac_list is not provided,"
                    + "maf_list is not provided, and z-scores are being used"
                )
                continue
            low_mac_mask = mac < single_population_mac_thresh
            XTX_list[i][low_mac_mask, :] = 0
            XTX_list[i][:, low_mac_mask] = 0
            XTY_list[i][low_mac_mask] = 0
            logger.info(
                f"Censored {np.sum(low_mac_mask)} variants in population {i}"
                + " due to low population-specific MAC"
            )

    # do minor allele frequency filtering across populations. this filter
    # only censors a variant if it has low MAF in all populations
    if multi_population_maf_thresh > 0:
        low_maf_across_populations_mask = np.ones(XTX_list[0].shape[0], dtype=bool)
        for i in range(K):
            n = population_sizes[i]
            if maf_list is not None:
                maf = np.minimum(maf_list[i], 1 - maf_list[i])
            elif mac_list is not None:
                maf = np.minimum(mac_list[i] / (2 * n), 1 - mac_list[i] / (2 * n))
            elif b_list is not None:
                var_x = np.minimum(np.diag(XTX_list[i]) / (n - 1), 0.5)
                maf = 1 / 2 - np.sqrt(1 - 2 * var_x) / 2
            else:
                logger.info(
                    "Skipping MAF filtering because mac_list is not provided,"
                    + "maf_list is not provided, and z-scores are being used"
                )
            low_maf_across_populations_mask = low_maf_across_populations_mask & (
                maf < multi_population_maf_thresh
            )
        logger.info(
            f"Censored {np.sum(low_maf_across_populations_mask)} variants in"
            + " all populations due to low MAF across all populations"
        )
        for i in range(K):
            XTX_list[i][low_maf_across_populations_mask, :] = 0
            XTX_list[i][:, low_maf_across_populations_mask] = 0
            XTY_list[i][low_maf_across_populations_mask] = 0

    if np.isscalar(rho):
        rho = np.ones((K, K), dtype=float_type) * rho  # type: ignore
        rho[np.diag_indices(K)] = 1  # type: ignore

    s = susie_multi_ss(
        XTX_list=XTX_list,
        XTY_list=XTY_list,
        YTY_list=YTY_list,
        rho=rho,
        population_sizes=population_sizes,
        L=L,
        scaled_prior_variance=scaled_prior_variance,
        prior_weights=prior_weights,
        standardize=standardize,
        pop_spec_standardization=pop_spec_standardization,
        estimate_residual_variance=estimate_residual_variance,
        estimate_prior_variance=estimate_prior_variance,
        estimate_prior_method=estimate_prior_method,
        prior_tol=prior_tol,
        max_iter=max_iter,
        tol=tol,
        verbose=verbose,
        iter_before_zeroing_effects=iter_before_zeroing_effects,
        pop_spec_effect_priors=pop_spec_effect_priors,
        R_list=R_list,
        coverage=coverage,
        min_abs_corr=min_abs_corr,
        float_type=float_type,
        low_memory_mode=low_memory_mode,
        recover_R=recover_R,
    )

    return s


susie_multi_rss = multisusie_rss


def recover_XTX_and_XTY(b, s, R, YTY, n):
    """
    Recover XTX and XTY from GWAS summary statistics.

    INPUT R IS MUTATED
    THIS FUNCTION MUTATES INPUT R to reduce memory consumpation. BE CAREFUL!!
    see page 4 of the supplement of Zou et al. 2022 PLoS Genetics for a derivation

    Parameters
    ----------
    b: length-P numpy array of floats representing effect sizes
    s: length-P numpy array of floats representing standard errors
    R: PxP numpy array of floats representing the LD correlation matrix
    YTY: float representing the sample variance of the outcome,
    n: integer representing the number of samples in the GWAS

    Returns
    -------
    R: PxP numpy array representing the LD correlation matrix, with low MAF/MAC
        variants censored. Note that this is the same object as the input R
    XTY: length-P numpy array representing the X^T Y vector
    n_censored: integer representing the number of variants censored
    """
    #  see page 4 of the supplement of Zou et al. 2022 PLoS Genetics for a derivation
    sigma2 = YTY / ((b / s) ** 2 + n - 2)
    XTY = np.nan_to_num(sigma2 * b / (s**2))
    dR = np.nan_to_num(sigma2 / (s**2))
    R *= np.expand_dims(np.sqrt(dR), axis=1)
    R *= np.sqrt(dR)
    np.nan_to_num(R, copy=False)

    return (R, XTY)


def recover_XTX_and_XTY_from_Z(z, R, n, float_type=np.float32):
    """
    Recover XTX and XTY from z scores and LD correlation matri.

    THIS FUNCTION MUTATES INPUT R to reduce memory consumpation. BE CAREFUL!!
    This is equivalent to using standardized genotype and phenotype

    Parameters
    ----------
    z: length-P numpy array of floats representing GWAS z scores
    R: PxP numpy array of floats representing the LD correlation matrix
    n: integer representing the number of samples in the GWAS

    Returns
    -------
    XTX: PxP numpy array representing the X^T X matrix
    XTY: length-P numpy array representing the X^T Y vector
    """
    adj = (n - 1) / (z**2 + n - 2)
    z = np.sqrt(adj) * z
    R *= n - 1
    XTY = (np.sqrt(n - 1) * z).astype(float_type)
    np.nan_to_num(XTY, copy=False)
    np.nan_to_num(R, copy=False)

    return R, XTY


def susie_multi_ss(
    XTX_list,
    XTY_list,
    YTY_list,
    rho,
    population_sizes,
    L=10,
    scaled_prior_variance=0.2,
    prior_weights=None,
    standardize=True,
    pop_spec_standardization=False,
    estimate_residual_variance=True,
    estimate_prior_variance=True,
    estimate_prior_method="early_EM",
    pop_spec_effect_priors=True,
    iter_before_zeroing_effects=5,
    prior_tol=1e-9,
    max_iter=100,
    tol=1e-3,
    verbose=False,
    R_list=None,
    coverage=0.95,
    min_abs_corr=0.5,
    float_type=np.float32,
    low_memory_mode=False,
    recover_R=False,
):
    r"""
    Run MultiSuSiE on sufficient statistics.

    This function runs MultiSuSiE on sufficient statistics. It is designed to
    be run from the top-level function multisusie_rss, but can be run directly.
    It will not censor based on MAF/MAC, unlike multisusie_rss.

    Parameters
    ----------
    XTX_list: length K list of PxP numpy arrays representing the X^T X matrices
    XTY_list: length K list of length P numpy arrays representing the X^T Y vectors
    YTY_list: length K list of floats representing the sample variance of the outcome
    rho: PxP numpy array representing the effect size correlation matrix
    population_sizes: list of integers representing the number of samples in
        the GWAS for each population
    L: integer representing the maximum number of causal variants
    scaled_prior_variance: float representing the effect size prior variance,
        scaled by the residual variance. It's fine to set this to a number
        larger than what you expect the squared effect size to be (like the
        default value of 0.2) as long as estimate_prior_variance is set to True
        and estimate_prior_method is not set to None.
    prior_weights: numpy P-array of floats representing the prior probability
        of causality for each variant. Give None to use a uniform prior
    standardize: boolean, whether to adjust summmary statistics to be as if
        genotypes were standardized to have mean 0 and variance 1
    pop_spec_standardization: boolean, if standardize is True, whether to
        adjust summary statistics to be as if genotypes were standardized
        separately for each population, or pooled and then standardized
    estimate_residual_variance: boolean, whether to estimate the residual variance,
        $\\sigma^2_k$ in the manuscript
    estimate_prior_variance: boolean, whether to estimate the prior variance,
        $A^{(l)}$ in the manuscript
    estimate_prior_method: string, method to estimate the prior variance. Recommended
        values are 'early_EM' or None
    pop_spec_effect_priors: boolean, whether to estimate separate prior
        variance parameters for each population
    iter_before_zeroing_effects: integer, number of iterations to run before
        zeroing out component-population pairs (or components if
        pop_spec_effect_priors is False) that have a lower likelihood than a
        null model
    prior_tol: float which places a filter on the minimum prior variance
        for a component to be included when estimating PIPs
    max_iter: integer, maximum number of iterations to run
    tol: float, after iter_before_zeroing_effects iterations, results
        are returned if the ELBO increases by less than tol in an ieration
    verbose: boolean which indicates if an progress bar should be displayed
    R_list: length K list of PxP numpy arrays representing the LD correlation.
        If set to None, and min_abs_corr > 0, the LD correlation matrices will
        be recovered from XTX_list.
    coverage: float representing the minimum coverage of credible sets
    min_abs_corr: float representing the minimum absolute correlation between
        any pair of variants in a credible set. For each pair of variants,
        the max is taken across ancestries. In the case where min_abs_corr = 0,
        low_memory_mode = True, and recover_R = False, the purity of credible
        sets will not be calculated.
    float_type: numpy float type used. Set to np.float32 to minimize memory
        consumption
    low_memory_mode: boolean, if True, the input R_list will be modified in place.
        Decreases memory consumption by a factor of two. BUT, THE INPUT R_list
        WILL BE OVERWRITTEN. If you need R_list, set low_memory_mode to False
    recover_R: boolean, if True, the R matrices will be recovered from XTX,
        BUT variants with MAC/MAF estimate less than mac_filter/maf_filter will
        be censored.

    Returns
    -------
    an object containing results with the following attributes:
        alpha: L x P numpy array of single-effect regression posterior
            inclusion probabilities
        mu: K x L x P numpy array of single-effect regression effect size
            posterior means, conditional on each variant being the causal variant
        mu2: K x K x L x P numpy array of single-effect regression effect size
            posterior seconds moments, conditional on each variant being the
            causal variant
        sigma2: length-K numpy array of residual variance estimates
        pi: length-P numpy array of prior inclusion probabilities
        n: length-K numpy array of sample sizes
        L: integer representing the maximum number of causal variants
        V: K x L numpy array of effect size prior variance estimates
        ER2: length-K numpy array of expected squared residuals
        KL: L x 1 numpy array of Kullback-Leibler divergences for each single
            effect regression
        lbf: L x 1 numpy array of log Bayes factors for each single effect
            regression
        converged: boolean indicating whether the algorithm converged
    """
    # check input dimensions
    assert len(XTX_list) == len(XTY_list)
    assert np.all([XTX.shape[1] == XTX_list[0].shape[1] for XTX in XTX_list])
    assert np.all(
        [XTX.shape[0] == XTY.shape[0] for (XTX, XTY) in zip(XTX_list, XTY_list)]
    )
    if prior_weights is not None:
        prior_weights = prior_weights.astype(float_type)

    assert not np.any([np.any(np.isnan(XTX)) for XTX in XTX_list])

    # compute w_pop (the relative size of each population)
    population_sizes = np.array(population_sizes)
    w_pop = (population_sizes / population_sizes.sum()).astype(float_type)

    # compute rho properties
    rho = rho.astype(float_type)
    logdet_rho_sign, logdet_rho = np.linalg.slogdet(rho)
    assert logdet_rho_sign > 0

    X_l2_arr = np.array([np.diag(XTX) for XTX in XTX_list], dtype=float_type)

    # calculate scaling factors
    if standardize:
        csd = np.sqrt(X_l2_arr / (np.expand_dims(population_sizes, axis=1) - 1))
        if not pop_spec_standardization:
            csd = csd.T.dot(w_pop)
            csd = csd * np.ones((len(XTX_list), csd.shape[0]), dtype=float_type)
        for pop_i in range(len(XTX_list)):
            XTX_list[pop_i] *= 1 / csd[pop_i, :]
            XTX_list[pop_i] *= 1 / np.expand_dims(csd[pop_i, :], 1)
            XTY_list[pop_i] = XTY_list[pop_i] / csd[pop_i, :]
        X_l2_arr = np.array([np.diag(XTX) for XTX in XTX_list], dtype=float_type)
    else:
        csd = np.ones((len(XTX_list), XTX_list[0].shape[1]), dtype=float_type)
    varY = np.array(
        [YTY / (n - 1) for (YTY, n) in zip(YTY_list, population_sizes)],
        dtype=float_type,
    )
    varY_pooled = (np.sum(YTY_list) / (np.sum(population_sizes) - 1)).astype(float_type)
    residual_variance = varY

    # init setup
    p = XTX_list[0].shape[1]

    if np.isscalar(scaled_prior_variance) & standardize:
        assert 0 < scaled_prior_variance <= 1
    if prior_weights is None:
        prior_weights = np.zeros(p, dtype=float_type) + 1.0 / p
    else:
        prior_weights = (prior_weights / np.sum(prior_weights)).astype(float_type)
    assert prior_weights.shape[0] == p
    if p < L:
        L = p

    # initialize s, which tracks the current model parameter estimates
    s = S(
        population_sizes,
        L,
        XTX_list,
        scaled_prior_variance,
        residual_variance,
        varY_pooled,
        prior_weights,
        float_type=float_type,
    )

    elbo = np.zeros(max_iter + 1) + np.nan
    elbo[0] = -np.inf
    check_null_threshold = 0.0

    # run the IBSS algorithm
    tqdm_iter = tqdm(list(range(max_iter)), disable=not verbose, file=sys.stdout)
    for i in tqdm_iter:
        tqdm_iter.set_description("iteration %d/%d" % (i + 1, max_iter))

        # set the prior variance estimation method for this iteration
        if estimate_prior_method == "early_EM":
            if i == 0:
                current_estimate_prior_method = None
            else:
                current_estimate_prior_method = "early_EM"
        elif (estimate_prior_method is not None) and (
            estimate_prior_method.split("_")[-1] == "init"
        ):
            if i == 0:
                current_estimate_prior_method = estimate_prior_method.split("_")[0]
            else:
                current_estimate_prior_method = "EM"
        else:
            current_estimate_prior_method = estimate_prior_method

        if i < iter_before_zeroing_effects:
            current_check_null_threshold = -np.inf
        else:
            current_check_null_threshold = check_null_threshold

        # update all L single effect regressions. s is modified in place.
        update_each_effect(
            XTX_list,
            XTY_list,
            s,
            X_l2_arr,
            w_pop,
            rho,
            estimate_prior_variance,
            current_estimate_prior_method,  # type: ignore
            check_null_threshold=current_check_null_threshold,
            pop_spec_effect_priors=pop_spec_effect_priors,
            float_type=float_type,
        )

        # update the ER2 parameter of s. Used by get_objective and
        # estimate_resdiaul_variance_func
        s.ER2 = update_ER2(XTX_list, XTY_list, YTY_list, s, X_l2_arr)

        # update the ELBO and check for convergence
        elbo[i + 1] = get_objective(XTX_list, XTY_list, s, YTY_list, X_l2_arr)
        if verbose:
            logger.info("objective: %s" % (elbo[i + 1]))

        if ((elbo[i + 1] - elbo[i]) < tol) and (i >= (iter_before_zeroing_effects + 1)):
            s.converged = True
            tqdm_iter.close()
            break

        if estimate_residual_variance:
            s.sigma2 = estimate_residual_variance_func(
                XTX_list,
                XTY_list,
                YTY_list,
                s,
                X_l2_arr,
                population_sizes,
                float_type=float_type,
            )
            if np.any(s.sigma2 < 0):

                logger.info(
                    "minimum resdiual variance less than 0. Is there mismatch between the correlation matrix and association statistics?"
                )

    elbo = elbo[1 : i + 2]  # Remove first (infinite) entry, and trailing NAs.
    s.elbo = elbo  # type: ignore
    s.niter = i + 1  # type: ignore

    if not s.converged:
        logger.warning("IBSS algorithm did not converge in %d iterations" % (max_iter))

    s.fitted = s.Xr_list  # type: ignore

    s.pip = susie_get_pip(s, prior_tol=prior_tol)  # type: ignore
    s.coef = np.array([np.squeeze(np.sum(s.mu[k] * s.alpha, axis=0) / csd[k, :]) for k in range(len(XTX_list))])  # type: ignore
    s.coef_sd = np.array(  # type: ignore
        [
            (
                np.squeeze(
                    np.sqrt(
                        np.sum(s.alpha * s.mu2[k, k] - (s.alpha * s.mu[k]) ** 2, axis=0)
                    )
                    / csd[k, :]
                )
            )
            for k in range(len(XTX_list))
        ]
    )

    if (low_memory_mode and (min_abs_corr > 0)) or (low_memory_mode and recover_R):
        for i in range(len(XTX_list)):
            recover_R_from_XTX(XTX_list[i], X_l2_arr[i])
        R_list = XTX_list

    s.sets = susie_get_cs(  # type: ignore
        s=s,
        R_list=R_list,
        coverage=coverage,
        min_abs_corr=min_abs_corr,
        dedup=True,
        n_purity=np.inf,  # type: ignore
        calculate_purity=(not low_memory_mode) or (min_abs_corr > 0) or recover_R,
    )

    return s


def update_each_effect(
    XTX_list,
    XTY_list,
    s,
    X_l2_arr,
    w_pop,
    rho,
    estimate_prior_variance=False,
    estimate_prior_method="optim",
    check_null_threshold=0.0,
    pop_spec_effect_priors=True,
    float_type=np.float32,
):
    """
    Update each single effect regression.

    This calculates updated single effect regression parameter estimates

    Parameters
    ----------
    XTX_list: length K list of PxP numpy arrays representing the X^T X matrices
    XTY_list: length K list of length P numpy arrays representing the X^T Y vectors
    s: S object representing the model parameter estimates prior to the update
    X_l2_arr: length K numpy array of floats representing the diagonal of X^T X
    w_pop: length K numpy array of floats representing the relative size of each population
    rho: PxP numpy array representing the effect size correlation matrix
    estimate_prior_variance: boolean, whether to estimate effect size prior variance
    estimate_prior_method: string, method to estimate the prior variance. Recommended
        values are 'early_EM' or None
    check_null_threshold: float representing the difference in loglikelihood that
        a component with a non-zero effect size prior variance must beat the null
        model by to be retained.
    pop_spec_effect_priors: boolean, whether to estimate separate prior effect
        size variances for each population
    float_type: numpy float type used. Set to np.float32 to minimize memory

    Returns
    -------
    Nothing. S is modified in place.
    """
    if not estimate_prior_variance:
        estimate_prior_method = None
    L = s.alpha.shape[0]
    num_pop = len(XTX_list)

    for effect_index in range(L):  # type: ignore
        R_list = []

        # add the estimate of the current effect back to the residualized
        # XTXY vector
        for k in range(len(XTX_list)):
            s.Xr_list[k] -= XTX_list[k].dot(
                s.alpha[effect_index] * s.mu[k, effect_index]
            )
            R_list.append(XTY_list[k] - s.Xr_list[k])

        # get the currrent single effect model parameter estimates after
        # residualizing all other effects
        res = single_effect_regression(
            R_list,
            XTX_list,
            s.V[:, effect_index],
            X_l2_arr,
            w_pop,
            rho,
            residual_variance=s.sigma2,
            prior_weights=s.pi,
            optimize_V=estimate_prior_method,
            check_null_threshold=check_null_threshold,  # type: ignore
            pop_spec_effect_priors=pop_spec_effect_priors,
            alpha=s.alpha[effect_index, :],
            mu2=s.mu2[:, :, effect_index, :],
            float_type=float_type,
        )

        # Update the variational estimate of the posterior mean.
        s.mu[:, effect_index, :] = res.mu
        s.alpha[effect_index, :] = res.alpha
        s.mu2[:, :, effect_index, :] = res.mu2
        s.V[:, effect_index] = res.V
        s.lbf[effect_index] = res.lbf_model
        s.KL[effect_index] = -res.lbf_model + SER_posterior_e_loglik(
            X_l2_arr,
            R_list,
            s.sigma2,
            res.mu * res.alpha,
            res.mu2[range(num_pop), range(num_pop)] * res.alpha,
        )
        for k in range(len(XTX_list)):
            s.Xr_list[k] += XTX_list[k].dot(
                s.alpha[effect_index] * s.mu[k, effect_index]
            )


def single_effect_regression(
    XTY_list,
    XTX_list,
    V,
    X_l2_arr,
    w_pop,
    rho,
    residual_variance,
    prior_weights=None,
    optimize_V=None,
    check_null_threshold=0,
    pop_spec_effect_priors=True,
    alpha=None,
    mu2=None,
    float_type=np.float32,
):
    """
    Fit a multi-population single effect regression model.

    Parameters
    ----------
    XTY_list: length K list of length P numpy arrays representing the X^T Y vectors
    XTX_list: length K list of PxP numpy arrays representing the X^T X matrices
    V: length K numpy array of floats representing the effect size prior variance
    X_l2_arr: length K numpy array of floats representing the diagonal of X^T X
    w_pop: length K numpy array of floats representing the relative size of each population
    rho: PxP numpy array representing the effect size correlation matrix
    residual_variance: length K numpy array of floats representing the residual
        variance in each population
    prior_weights: length P numpy array of floats representing the prior probability
        of causality for each variant
    optimize_V: string representing the method to use to optimize the effect size
    check_null_threshold: float representing the difference in loglikelihood that
        a component with a non-zero effect size prior variance must beat the null
        model by to be retained.
    pop_spec_effect_priors: boolean, whether to estimate separate prior effect
        size variances for each population
    alpha: length P numpy array of floats representing the current posterior
        inclusion probability estimates
    mu2: K x K x P numpy array of floats representing the current posterior
        second moment estimates
    float_type: numpy float type used. Set to np.float32 to minimize memory

    Returns
    -------
    an object containing results with the following attributes:
        alpha: length P numpy array of single effect regression posterior
            inclusion probabilities
        mu: K x P numpy array of single effect regression effect sizes conditional
            on each variant being the causal variant
        mu2: K x K x P numpy array of single effect regression effect size second
            moments conditional on each variant being the causal variant
        lbf: K x P numpy array representing the log Bayes factor for each
            variant being the causal variant
        lbf_model: float representing the log Bayes factor for the model,
            aggregating over variants
        V: length K numpy array of floats representing the effect size prior
    """
    compute_lbf_params = (
        XTY_list,
        XTX_list,
        X_l2_arr,
        rho,
        residual_variance,
        False,
        float_type,
    )

    if optimize_V not in ["EM", "EM_corrected", None]:
        V = optimize_prior_variance(
            optimize_V,  # type: ignore
            prior_weights,  # type: ignore
            rho.shape[0],
            compute_lbf_params=compute_lbf_params,
            alpha=alpha,  # type: ignore
            post_mean2=mu2,  # type: ignore
            w_pop=w_pop,
            check_null_threshold=check_null_threshold,
            pop_spec_effect_priors=pop_spec_effect_priors,
            current_V=V,
            float_type=float_type,
        )

    # compute the log Bayes factor for each variant being the causal variant
    # and posterior mean estimates for each variant  under the assumption
    # that it is the causal variant
    lbf, post_mean, post_mean2 = compute_lbf(
        V=V,  # type: ignore
        XTY_list=XTY_list,
        XTX_list=XTX_list,
        X_l2_arr=X_l2_arr,
        rho=rho,
        residual_variance=residual_variance,
        return_moments=True,
        float_type=float_type,
    )

    # compute posterior inclusion probabilities (not combined over single effect regressions)
    maxlbf = np.max(lbf)
    w = np.exp(lbf - maxlbf)
    w_weighted = w * prior_weights
    weighted_sum_w = w_weighted.sum()
    alpha = w_weighted / weighted_sum_w
    lbf_model = maxlbf + np.log(weighted_sum_w)

    if optimize_V in ["EM", "EM_corrected"]:
        V = optimize_prior_variance(
            optimize_V,  # type: ignore
            prior_weights,  # type: ignore
            rho.shape[0],
            compute_lbf_params=compute_lbf_params,
            alpha=alpha,
            post_mean2=post_mean2,
            w_pop=w_pop,
            check_null_threshold=check_null_threshold,
            pop_spec_effect_priors=pop_spec_effect_priors,
            float_type=float_type,
        )

    res = SER_RESULTS(
        alpha=alpha, mu=post_mean, mu2=post_mean2, lbf=lbf, lbf_model=lbf_model, V=V
    )

    return res


def loglik(
    V: np.ndarray, prior_weights: np.ndarray, compute_lbf_params: tuple
) -> float:
    """
    Calculate the log-likelihood of the model.

    This function computes the log-likelihood of the model given the effect size prior variance,
    prior weights, and parameters for computing the log Bayes factor.

    Parameters
    ----------
    V : np.ndarray
        Array of floats representing the effect size prior variance for each population.
    prior_weights : np.ndarray
        Array of floats representing the prior probability of causality for each variant.
    compute_lbf_params : tuple
        Tuple containing parameters required to compute the log Bayes factor.

    Returns
    -------
    float
        The log-likelihood of the model.
    """
    lbf = compute_lbf(V, *compute_lbf_params)
    maxlbf = np.max(lbf)
    w = np.exp(lbf - maxlbf)
    w_weighted = w * prior_weights
    weighted_sum_w = w_weighted.sum()
    loglik = maxlbf + np.log(weighted_sum_w)
    return loglik


def optimize_prior_variance(
    optimize_V: str,
    prior_weights: np.ndarray,
    num_pops: int,
    compute_lbf_params: tuple = None,  # type: ignore
    alpha: np.ndarray = None,  # type: ignore
    post_mean2: np.ndarray = None,  # type: ignore
    w_pop: np.ndarray = None,  # type: ignore
    check_null_threshold: float = 0,
    pop_spec_effect_priors: bool = False,
    current_V: np.ndarray = None,  # type: ignore
    float_type: type = np.float32,
    loglik_function: callable = loglik,  # type: ignore
) -> np.ndarray:
    """
    Optimize the prior variance for the single effect regression model.

    This function optimizes the prior variance for the single effect regression model using
    different methods such as 'optim', 'EM', 'early_EM', and 'grid'.

    Parameters
    ----------
    optimize_V : str
        Method to optimize the prior variance. Options are 'optim', 'EM', 'early_EM', and 'grid'.
    prior_weights : np.ndarray
        Array of floats representing the prior probability of causality for each variant.
    num_pops : int
        Number of populations.
    compute_lbf_params : tuple, optional
        Tuple containing parameters required to compute the log Bayes factor.
    alpha : np.ndarray, optional
        Array of floats representing the current posterior inclusion probability estimates.
    post_mean2 : np.ndarray, optional
        Array of floats representing the current posterior second moment estimates.
    w_pop : np.ndarray, optional
        Array of floats representing the relative size of each population.
    check_null_threshold : float, optional
        Difference in loglikelihood that a component with a non-zero effect size prior variance
        must beat the null model by to be retained.
    pop_spec_effect_priors : bool, optional
        Whether to estimate separate prior effect size variances for each population.
    current_V : np.ndarray, optional
        Array of floats representing the current prior variance estimates.
    float_type : type, optional
        Numpy float type used for calculations (default is np.float32).
    loglik_function : callable, optional
        Function to calculate the log-likelihood of the model (default is loglik).

    Returns
    -------
    np.ndarray
        Array of floats representing the optimized prior variance for each population.
    """
    if optimize_V == "optim":
        if pop_spec_effect_priors:
            raise Exception(
                'estimate_prior_method="optim" with '
                + "pop_spec_effect_priors=True has not been implemented"
            )
        else:

            def neg_loglik_logscale(lV):
                return -loglik_function(
                    np.array([np.exp(lV) for i in range(num_pops)]),
                    prior_weights,
                    compute_lbf_params,
                )

            opt_obj = minimize_scalar(neg_loglik_logscale, bounds=(-30, 15))
            lV = opt_obj.x  # type: ignore
            V = np.exp(lV)
    elif optimize_V in ["EM", "early_EM"]:
        V = np.array(
            [np.sum(alpha * post_mean2[i, i]) for i in range(num_pops)],
            dtype=float_type,
        )
        if not pop_spec_effect_priors:
            V = (w_pop.dot(V)).astype(float_type)
    elif optimize_V == "grid":
        if pop_spec_effect_priors:
            raise Exception(
                'estimate_prior_method="grid" with '
                + "pop_spec_effect_priors=True has not been implemented"
            )
        else:
            V_arr = np.logspace(-7, -1, 13)
            llik_arr = np.array(
                [loglik_function(V, prior_weights, compute_lbf_params) for V in V_arr]
            )
            V = V_arr[np.argmax(llik_arr)]
    else:
        raise ValueError("unknown optimization method")

    if not pop_spec_effect_priors:
        V = V * np.ones(post_mean2.shape[0], dtype=float_type)
        # set V exactly 0 if that beats the numerical value by check_null_threshold in loglik.
        delta_loglik = (
            loglik_function(0, prior_weights, compute_lbf_params)
            + check_null_threshold
            - loglik_function(V, prior_weights, compute_lbf_params)
        )
        if np.isclose(delta_loglik, 0) or delta_loglik >= 0:
            V = 0
    else:
        if check_null_threshold == (-1 * np.inf):
            return V
        elif np.all(np.isclose(V, 0)):
            return 0  # type: ignore
        # Compare our current effect prior to null models
        else:
            V_list = [V]
            # zero out each population, one at a time
            for i in range(num_pops):
                if not np.isclose(V[i], 0):
                    V_copy = V.copy()
                    V_copy[i] = 0
                    V_list.append(V_copy)
            V_list.append(np.array([np.zeros(num_pops, dtype=float_type)]))
        llik_arr = np.array(
            [
                loglik_function(np.array(V), prior_weights, compute_lbf_params)
                for V in V_list
            ]
        )
        llik_arr = llik_arr + np.array(
            [0] + [check_null_threshold for i in range(len(V_list) - 1)]
        )
        V = V_list[np.argmax(llik_arr)]
    if isinstance(V, np.ndarray):
        V[V < 0] = 0
    elif V < 0:
        V = 0

    return V  # type: ignore


def compute_lbf(
    V: np.ndarray,
    XTY_list: list,
    XTX_list: list,
    X_l2_arr: np.ndarray,
    rho: np.ndarray,
    residual_variance: np.ndarray,
    return_moments: bool = False,
    float_type: type = np.float32,
) -> tuple:
    """
    Compute the log Bayes factor (LBF) and optionally posterior moments.

    This function calculates the log Bayes factor for each variant using the provided
    effect size prior variance, X^T Y vectors, X^T X diagonal elements, effect size
    correlation matrix, and residual variance. Optionally, it can also return the
    posterior means and second moments.

    Parameters
    ----------
    V : np.ndarray
        Array of floats representing the effect size prior variance for each population.
    XTY_list : list
        List of length-P numpy arrays representing the X^T Y vectors for each population.
    XTX_list : list
        List of PxP numpy arrays representing the X^T X matrices for each population.
    X_l2_arr : np.ndarray
        Array of floats representing the diagonal of X^T X for each population.
    rho : np.ndarray
        2D array of floats representing the effect size correlation matrix.
    residual_variance : np.ndarray
        Array of floats representing the residual variance for each population.
    return_moments : bool, optional
        Whether to return posterior moments (default is False).
    float_type : type, optional
        Numpy float type used for calculations (default is np.float32).

    Returns
    -------
    tuple
        If return_moments is True, returns a tuple containing the log Bayes factor (LBF),
        posterior means, and posterior second moments. Otherwise, returns only the LBF.
    """
    num_variables = XTX_list[0].shape[0]
    num_pops = len(XTX_list)

    if np.all(np.isclose(V, 0)):
        lbf = np.zeros(num_variables, dtype=float_type)
        if return_moments:
            post_mean = np.zeros((num_pops, num_variables), dtype=float_type)
            post_mean2 = np.zeros((num_pops, num_pops, num_variables), dtype=float_type)

    elif np.any(np.isclose(V, 0)):
        nonzero_pops = np.flatnonzero(~np.isclose(V, 0))
        lbf_out = compute_lbf(
            V=V[nonzero_pops],
            XTY_list=[XTY_list[i] for i in nonzero_pops],
            XTX_list=[XTX_list[i] for i in nonzero_pops],
            X_l2_arr=X_l2_arr[nonzero_pops],
            rho=rho[nonzero_pops, :][:, nonzero_pops],
            residual_variance=residual_variance[nonzero_pops],
            return_moments=return_moments,
            float_type=float_type,
        )
        if return_moments:
            post_mean = np.zeros((num_pops, num_variables), dtype=float_type)
            post_mean2 = np.zeros((num_pops, num_pops, num_variables), dtype=float_type)
            lbf = lbf_out[0]
            post_mean[nonzero_pops] = lbf_out[1]
            post_mean2[np.ix_(nonzero_pops, nonzero_pops)] = lbf_out[2]
        else:
            lbf = lbf_out

    else:
        XTY = np.ascontiguousarray(np.stack(XTY_list, axis=1))
        if return_moments:
            try:
                lbf, post_mean, post_mean2 = compute_lbf_and_moments(
                    V=V,
                    XTY=XTY,
                    X_l2_arr=X_l2_arr,
                    rho=rho,
                    residual_variance=residual_variance,
                    float_type=float_type,
                )
            except Exception:
                lbf, post_mean, post_mean2 = compute_lbf_and_moments_safe(
                    V=V,
                    XTY=XTY,
                    X_l2_arr=X_l2_arr,
                    rho=rho,
                    residual_variance=residual_variance,
                    float_type=float_type,
                )
        else:
            lbf = compute_lbf_no_moments(
                V=V,
                XTY=XTY,
                X_l2_arr=X_l2_arr,
                rho=rho,
                residual_variance=residual_variance,
                float_type=float_type,
            )
    if return_moments:
        return lbf, post_mean, post_mean2
    else:
        return lbf  # type: ignore


@numba.jit(nopython=True, cache=False)
def compute_lbf_no_moments(
    V: np.ndarray,
    XTY: np.ndarray,
    X_l2_arr: np.ndarray,
    rho: np.ndarray,
    residual_variance: np.ndarray,
    float_type: type = np.float32,
) -> np.ndarray:
    """
    Compute the log Bayes factor (LBF) without posterior moments.

    This function calculates the log Bayes factor for each variant using the provided
    effect size prior variance, X^T Y vectors, X^T X diagonal elements, effect size
    correlation matrix, and residual variance.

    Parameters
    ----------
    V : np.ndarray
        Array of floats representing the effect size prior variance for each population.
    XTY : np.ndarray
        2D array of floats representing the X^T Y vectors for each population.
    X_l2_arr : np.ndarray
        Array of floats representing the diagonal of X^T X for each population.
    rho : np.ndarray
        2D array of floats representing the effect size correlation matrix.
    residual_variance : np.ndarray
        Array of floats representing the residual variance for each population.
    float_type : type, optional
        Numpy float type used for calculations (default is np.float32).

    Returns
    -------
    np.ndarray
        Array of floats representing the log Bayes factor for each variant.
    """
    num_pops = XTY.shape[1]
    num_variables = XTY.shape[0]

    lbf = np.zeros(num_variables, dtype=float_type)

    YT_invD_Z = XTY / residual_variance

    # compute a (the effects covariance matrix) and its inverse and log-determinant
    A = rho * np.sqrt(np.outer(V, V))
    inv_A = np.linalg.inv(A)
    logdet_A_sign, logdetA = np.linalg.slogdet(A)

    for i in range(num_variables):

        # compute the diagonal of q = z.t * inv(d) * z (this is a diagonal matrix)
        # this is (n_pop)
        Q_diag = X_l2_arr[:, i] / residual_variance

        # compute log-determinent for inv(a)+q
        # this is (n_pop, n_pop)
        Ainv_plus_Q = inv_A + np.diag(Q_diag)
        logdet_Ainv_plus_Q_sign, logdet_Ainv_plus_Q = np.linalg.slogdet(Ainv_plus_Q)
        assert logdet_Ainv_plus_Q_sign > 0

        # compute inv_ainv_plus_q_times_zt_invd_y
        inv_Ainv_plus_Q_times_ZT_invD_Y = np.linalg.solve(Ainv_plus_Q, YT_invD_Z[i, :])

        # compute log-bf for this variable
        lbf_1 = 0.5 * YT_invD_Z[i, :].dot(inv_Ainv_plus_Q_times_ZT_invD_Y)
        lbf_2 = -0.5 * (logdetA + logdet_Ainv_plus_Q)
        lbf[i] = lbf_1 + lbf_2

    return lbf


@numba.jit(nopython=True, cache=False)
def compute_lbf_and_moments(
    V: np.ndarray,
    XTY: np.ndarray,
    X_l2_arr: np.ndarray,
    rho: np.ndarray,
    residual_variance: np.ndarray,
    float_type: type = np.float32,
) -> tuple:
    """
    Compute the log Bayes factor and posterior moments.

    This function computes the log Bayes factor (LBF) and posterior moments for each variant
    using the provided effect size prior variance, X^T Y vectors, X^T X diagonal elements,
    effect size correlation matrix, and residual variance.

    Parameters
    ----------
    V : np.ndarray
        Array of floats representing the effect size prior variance for each population.
    XTY : np.ndarray
        2D array of floats representing the X^T Y vectors for each population.
    X_l2_arr : np.ndarray
        Array of floats representing the diagonal of X^T X for each population.
    rho : np.ndarray
        2D array of floats representing the effect size correlation matrix.
    residual_variance : np.ndarray
        Array of floats representing the residual variance for each population.
    float_type : type, optional
        Numpy float type used for calculations (default is np.float32).

    Returns
    -------
    tuple
        A tuple containing the log Bayes factor (LBF), posterior means, and posterior second moments.
    """
    num_pops = XTY.shape[1]
    num_variables = XTY.shape[0]

    lbf = np.zeros(num_variables, dtype=float_type)
    post_mean = np.zeros((num_pops, num_variables), dtype=float_type)
    post_mean2 = np.zeros((num_pops, num_pops, num_variables), dtype=float_type)

    YT_invD_Z = XTY / residual_variance

    # compute a (the effects covariance matrix) and its inverse and log-determinant
    A = rho * np.sqrt(np.outer(V, V))
    inv_A = np.linalg.inv(A)
    logdet_A_sign, logdetA = np.linalg.slogdet(A)

    for i in range(num_variables):

        # compute the diagonal of q = z.t * inv(d) * z (this is a diagonal matrix)
        Q_diag = X_l2_arr[:, i] / residual_variance

        # compute log-determinent for inv(a)+q
        Ainv_plus_Q = inv_A + np.diag(Q_diag)
        logdet_Ainv_plus_Q_sign, logdet_Ainv_plus_Q = np.linalg.slogdet(Ainv_plus_Q)
        assert logdet_Ainv_plus_Q_sign > 0

        # compute inv_ainv_plus_q_times_zt_invd_y
        inv_Ainv_plus_Q_times_ZT_invD_Y = np.linalg.solve(Ainv_plus_Q, YT_invD_Z[i, :])

        # compute log-bf for this variable
        lbf_1 = 0.5 * YT_invD_Z[i, :].dot(inv_Ainv_plus_Q_times_ZT_invD_Y)
        lbf_2 = -0.5 * (logdetA + logdet_Ainv_plus_Q)
        lbf[i] = lbf_1 + lbf_2

        # compute posterior moments for this variable
        AQ = A * Q_diag
        post_mean[:, i] = A.dot(YT_invD_Z[i, :]) - AQ.dot(
            inv_Ainv_plus_Q_times_ZT_invD_Y
        )
        post_covar_i = A - AQ.dot(A) + AQ.dot(np.linalg.solve(Ainv_plus_Q, AQ.T))
        post_mean2[:, :, i] = np.maximum(
            post_covar_i + np.outer(post_mean[:, i], post_mean[:, i]), 0
        )

    return lbf, post_mean, post_mean2


def compute_lbf_and_moments_safe(
    V: np.ndarray,
    XTY: np.ndarray,
    X_l2_arr: np.ndarray,
    rho: np.ndarray,
    residual_variance: np.ndarray,
    float_type: type = np.float32,
) -> tuple:
    """
    Compute the log Bayes factor and posterior moments safely.

    This function computes the log Bayes factor (LBF) and posterior moments for each variant
    using a more numerically stable approach. It is used when the standard method encounters
    numerical issues.

    Parameters
    ----------
    V : np.ndarray
        Array of floats representing the effect size prior variance for each population.
    XTY : np.ndarray
        2D array of floats representing the X^T Y vectors for each population.
    X_l2_arr : np.ndarray
        Array of floats representing the diagonal of X^T X for each population.
    rho : np.ndarray
        2D array of floats representing the effect size correlation matrix.
    residual_variance : np.ndarray
        Array of floats representing the residual variance for each population.
    float_type : type, optional
        Numpy float type used for calculations (default is np.float32).

    Returns
    -------
    tuple
        A tuple containing the log Bayes factor (LBF), posterior means, and posterior second moments.
    """
    num_pops = XTY.shape[1]
    num_variables = XTY.shape[0]

    lbf = np.zeros(num_variables, dtype=float_type)
    post_mean = np.zeros((num_pops, num_variables), dtype=float_type)
    post_mean2 = np.zeros((num_pops, num_pops, num_variables), dtype=float_type)

    YT_invD_Z = XTY / residual_variance

    # compute a (the effects covariance matrix) and its inverse and log-determinant
    A = rho * np.sqrt(np.outer(V, V))
    inv_A = np.linalg.inv(A)
    logdet_A_sign, logdetA = np.linalg.slogdet(A)

    for i in range(num_variables):

        # compute the diagonal of q = z.t * inv(d) * z (this is a diagonal matrix)
        Q_diag = X_l2_arr[:, i] / residual_variance

        # compute log-determinent for inv(a)+q
        Ainv_plus_Q = inv_A + np.diag(Q_diag)
        logdet_Ainv_plus_Q_sign, logdet_Ainv_plus_Q = np.linalg.slogdet(Ainv_plus_Q)
        assert logdet_Ainv_plus_Q_sign > 0

        # compute inv_ainv_plus_q_times_zt_invd_y
        inv_Ainv_plus_Q_times_ZT_invD_Y = np.linalg.solve(Ainv_plus_Q, YT_invD_Z[i, :])

        # compute log-bf for this variable
        lbf_1 = 0.5 * YT_invD_Z[i, :].dot(inv_Ainv_plus_Q_times_ZT_invD_Y)
        lbf_2 = -0.5 * (logdetA + logdet_Ainv_plus_Q)
        lbf[i] = lbf_1 + lbf_2

        # compute posterior moments for this variable
        AQ = A * Q_diag
        post_mean[:, i] = A.dot(YT_invD_Z[i, :]) - AQ.dot(
            inv_Ainv_plus_Q_times_ZT_invD_Y
        )
        post_covar_i = A - AQ.dot(A) + AQ.dot(np.linalg.solve(Ainv_plus_Q, AQ.T))
        post_mean2[:, :, i] = post_covar_i + np.outer(post_mean[:, i], post_mean[:, i])

    return lbf, post_mean, post_mean2


def get_objective(
    XTX_list: list, XTY_list: list, s: S, YTY_list: list, X_l2_arr: np.ndarray
) -> float:
    """
    Calculate the objective function value.

    This function computes the objective function value, which is the difference between the expected log-likelihood
    and the sum of Kullback-Leibler divergences for each single effect regression.

    Parameters
    ----------
    XTX_list : list of np.ndarray
        List of PxP numpy arrays representing the X^T X matrices for each population.
    XTY_list : list of np.ndarray
        List of length P numpy arrays representing the X^T Y vectors for each population.
    s : S
        An object containing the current model parameter estimates.
    YTY_list : list of float
        List of floats representing the sample variance of the outcome for each population.
    X_l2_arr : np.ndarray
        Array of floats representing the diagonal of X^T X for each population.

    Returns
    -------
    float
        The value of the objective function.
    """
    return Eloglik(XTX_list, XTY_list, s, YTY_list, X_l2_arr) - np.sum(s.KL)  # type: ignore


def Eloglik(
    XTX_list: list, XTY_list: list, s: S, YTY_list: list, X_l2_arr: np.ndarray
) -> float:
    """
    Calculate the expected log-likelihood of the model.

    This function computes the expected log-likelihood of the model given the sufficient statistics
    and model parameters.

    Parameters
    ----------
    XTX_list : list of np.ndarray
        List of PxP numpy arrays representing the X^T X matrices for each population.
    XTY_list : list of np.ndarray
        List of length P numpy arrays representing the X^T Y vectors for each population.
    s : S
        An object containing the current model parameter estimates.
    YTY_list : list of float
        List of floats representing the sample variance of the outcome for each population.
    X_l2_arr : np.ndarray
        Array of floats representing the diagonal of X^T X for each population.

    Returns
    -------
    float
        The expected log-likelihood of the model.
    """
    result = -0.5 * s.n.dot(np.log(2 * np.pi * s.sigma2))
    for i in range(len(XTX_list)):
        result -= 0.5 / s.sigma2[i] * s.ER2[i]
    return result  # type: ignore


def update_ER2(
    XTX_list: list, XTY_list: list, YTY_list: list, s: S, X_l2_arr: np.ndarray
) -> np.ndarray:
    """
    Update the expected residual sum of squares (ER2) for each population.

    This function calculates the expected residual sum of squares (ER2) for each population
    based on the provided sufficient statistics and model parameters.

    Parameters
    ----------
    XTX_list : list of np.ndarray
        List of PxP numpy arrays representing the X^T X matrices for each population.
    XTY_list : list of np.ndarray
        List of length P numpy arrays representing the X^T Y vectors for each population.
    YTY_list : list of float
        List of floats representing the sample variance of the outcome for each population.
    s : S
        An object containing the current model parameter estimates.
    X_l2_arr : np.ndarray
        Array of floats representing the diagonal of X^T X for each population.

    Returns
    -------
    np.ndarray
        Array of floats representing the updated expected residual sum of squares (ER2) for each population.
    """
    ER2 = np.zeros(len(XTX_list))
    for i in range(len(XTX_list)):
        ER2[i] = get_ER2(
            XTX_list[i],
            XTY_list[i],
            YTY_list[i],
            s.alpha,
            s.mu[i],
            s.mu2[i, i],
            X_l2_arr[i],
        )
    return ER2


def get_ER2(
    XTX: np.ndarray,
    XTY: np.ndarray,
    YTY: float,
    alpha: np.ndarray,
    mu: np.ndarray,
    mu2: np.ndarray,
    X_l2: np.ndarray,
) -> float:
    """
    Calculate the expected residual sum of squares (ER2).

    This function computes the expected residual sum of squares (ER2) for a given set of parameters.

    Parameters
    ----------
    XTX : np.ndarray
        The X^T X matrix.
    XTY : np.ndarray
        The X^T Y vector.
    YTY : float
        The Y^T Y scalar.
    alpha : np.ndarray
        The posterior inclusion probabilities.
    mu : np.ndarray
        The posterior means.
    mu2 : np.ndarray
        The posterior second moments.
    X_l2 : np.ndarray
        The diagonal elements of X^T X.

    Returns
    -------
    float
        The expected residual sum of squares (ER2).
    """
    B = alpha * mu  # beta should be lxp
    XB2 = np.sum(B.T * np.dot(XTX, B.T))
    betabar = np.sum(B, axis=0)
    postb2 = alpha * mu2

    result = (
        YTY
        - 2 * betabar.dot(XTY)
        + betabar.dot(np.dot(XTX, betabar))
        - XB2
        + np.sum(X_l2 * postb2)
    )
    return result


def SER_posterior_e_loglik(
    X_l2_arr: np.ndarray,
    XTY_list: list,
    s2: np.ndarray,
    Eb: np.ndarray,
    Eb2: np.ndarray,
) -> float:
    """
    Calculate the expected log-likelihood of the single effect regression posterior.

    This function computes the expected log-likelihood of the single effect regression posterior
    given the sufficient statistics and model parameters.

    Parameters
    ----------
    X_l2_arr : np.ndarray
        Array of floats representing the diagonal of X^T X for each population.
    XTY_list : list of np.ndarray
        List of length P numpy arrays representing the X^T Y vectors for each population.
    s2 : np.ndarray
        Array of floats representing the residual variance for each population.
    Eb : np.ndarray
        Array of floats representing the posterior mean effect sizes for each population.
    Eb2 : np.ndarray
        Array of floats representing the posterior second moments for each population.

    Returns
    -------
    float
        The expected log-likelihood of the single effect regression posterior.
    """
    result = 0
    for i in range(len(X_l2_arr)):
        result -= 0.5 / s2[i] * (-2 * XTY_list[i].dot(Eb[i]) + X_l2_arr[i].dot(Eb2[i]))
    return result


def estimate_residual_variance_func(
    XTX_list, XTY_list, YTY_list, s, X_l2_arr, population_sizes, float_type=np.float32
):
    """
    Estimate the residual variance for each population.

    This function calculates the residual variance for each population based on the
    provided sufficient statistics and model parameters.

    Parameters
    ----------
    XTX_list : list of np.ndarray
        List of PxP numpy arrays representing the X^T X matrices for each population.
    XTY_list : list of np.ndarray
        List of length P numpy arrays representing the X^T Y vectors for each population.
    YTY_list : list of float
        List of floats representing the sample variance of the outcome for each population.
    s : S
        An object containing the current model parameter estimates.
    X_l2_arr : np.ndarray
        Array of floats representing the diagonal of X^T X for each population.
    population_sizes : np.ndarray
        Array of integers representing the number of samples in the GWAS for each population.
    float_type : type, optional
        Numpy float type used for calculations (default is np.float32).

    Returns
    -------
    np.ndarray
        Array of floats representing the estimated residual variance for each population.
    """
    sigma2_arr = np.zeros(len(XTX_list), dtype=float_type)
    for i in range(len(XTX_list)):
        sigma2_arr[i] = s.ER2[i] / population_sizes[i]
    return sigma2_arr


def susie_get_pip(s, prior_tol=1e-9):
    """
    Calculate the posterior inclusion probabilities (PIPs) for each variant.

    This function computes the PIPs for each variant by aggregating the posterior
    inclusion probabilities across all components.

    Parameters
    ----------
    s : object
        An object containing the results of the MultiSuSiE model.
    prior_tol : float, optional
        Tolerance level for the prior variance to include a component (default is 1e-9).

    Returns
    -------
    np.ndarray
        Array of PIPs for each variant.
    """
    include_idx = np.any(s.V > prior_tol, axis=0)
    if not np.any(include_idx):
        return np.zeros(s.alpha.shape[1])
    res = s.alpha[include_idx, :]
    pips = 1 - np.prod(1 - res, axis=0)
    return pips


def susie_get_cs(
    s,
    R_list,
    coverage=0.95,
    min_abs_corr=0.5,
    dedup=True,
    n_purity=100,
    calculate_purity=True,
    X_list=None,
):
    """
    Identify credible sets and calculate their purity.

    This function identifies credible sets for each component in the model and calculates their purity.
    It also deduplicates the credible sets if specified.

    Parameters
    ----------
    s : object
        An object containing the results of the MultiSuSiE model.
    R_list : list of np.ndarray
        List of LD correlation matrices for each population.
    coverage : float, optional
        Desired coverage level for the credible sets (default is 0.95).
    min_abs_corr : float, optional
        Minimum absolute correlation threshold for credible sets (default is 0.5).
    dedup : bool, optional
        Whether to deduplicate the credible sets (default is True).
    n_purity : int, optional
        Maximum number of variants to consider for purity calculation (default is 100).
    calculate_purity : bool, optional
        Whether to calculate the purity of the credible sets (default is True).
    X_list : list of np.ndarray, optional
        List of genotype matrices for each population.

    Returns
    -------
    tuple
        A tuple containing the credible sets, their purity, claimed coverage, and inclusion mask.
    """
    if len(s.V.shape) == 2:
        include_mask = np.any(s.V > 1e-9, axis=0)
    else:
        include_mask = s.V > 1e-9

    status = in_CS(s.alpha, coverage)
    cs = [np.argwhere(status[i, :] == 1).flatten() for i in range(status.shape[0])]
    cs = [cs[i] if include_mask[i] else [] for i in range(len(cs))]
    claimed_coverage = np.array([np.sum(s.alpha[i, cs[i]]) for i in range(len(cs))])
    include_mask = include_mask & [len(csi) > 0 for csi in cs]

    if dedup:
        cs_set = set()
        for i in range(len(cs)):
            if tuple(cs[i]) in cs_set:
                include_mask[i] = False
            elif include_mask[i]:
                cs_set.add(tuple(cs[i]))
    if not np.any(include_mask):
        return ([[] for i in range(len(include_mask))], None, None, include_mask)

    if calculate_purity:
        purity = np.array(
            [
                (
                    get_purity_x(cs[i], R_list, min_abs_corr, n_purity, X_list)
                    if include_mask[i]
                    else np.nan
                )
                for i in range(len(cs))
            ]
        )
        include_mask[purity < min_abs_corr] = False
    else:
        purity = np.array([np.nan for i in range(len(cs))])

    return (cs, purity, claimed_coverage, include_mask)


def in_CS(alpha: np.ndarray, coverage: float) -> np.ndarray:
    """
    Determine which variants are included in the credible set for each component.

    This function calculates the credible set for each component by including variants
    until the cumulative sum of their posterior inclusion probabilities (PIPs) reaches or exceeds
    the specified coverage.

    Parameters
    ----------
    alpha : np.ndarray
        2D array of posterior inclusion probabilities (PIPs) for each variant and component.
    coverage : float
        Desired coverage level for the credible sets.

    Returns
    -------
    np.ndarray
        2D array indicating which variants are included in the credible set for each component (1 if included, 0 otherwise).
    """
    return np.apply_along_axis(in_CS_x, 1, alpha, coverage)


def in_CS_x(alpha: np.ndarray, coverage: float) -> np.ndarray:
    """
    Determine which variants are included in the credible set for a single component.

    This function calculates the credible set for a single component by including variants
    until the cumulative sum of their posterior inclusion probabilities (PIPs) reaches or exceeds
    the specified coverage.

    Parameters
    ----------
    alpha : np.ndarray
        Array of posterior inclusion probabilities (PIPs) for each variant.
    coverage : float
        Desired coverage level for the credible set.

    Returns
    -------
    np.ndarray
        Array indicating which variants are included in the credible set (1 if included, 0 otherwise).
    """
    n = n_in_CS_x(alpha, coverage)
    o = np.argsort(alpha)[::-1]
    result = np.zeros(alpha.shape, dtype=np.int32)
    result[o[:n]] = 1
    return result


def n_in_CS(alpha: np.ndarray, coverage: float) -> np.ndarray:
    """
    Calculate the number of variants to include in credible sets for multiple components.

    This function determines the number of variants to include in credible sets for each component
    such that the cumulative sum of their posterior inclusion probabilities (PIPs) reaches or exceeds
    the specified coverage.

    Parameters
    ----------
    alpha : np.ndarray
        2D array of posterior inclusion probabilities (PIPs) for each variant and component.
    coverage : float
        Desired coverage level for the credible sets.

    Returns
    -------
    np.ndarray
        Array of integers representing the number of variants to include in the credible set for each component.
    """
    return np.apply_along_axis(n_in_CS_x, 1, alpha, coverage)


def n_in_CS_x(alpha: np.ndarray, coverage: float) -> int:
    """
    Calculate the number of variants to include in a credible set.

    This function determines the number of variants to include in a credible set
    such that the cumulative sum of their posterior inclusion probabilities (PIPs)
    reaches or exceeds the specified coverage.

    Parameters
    ----------
    alpha : np.ndarray
        Array of posterior inclusion probabilities (PIPs) for each variant.
    coverage : float
        Desired coverage level for the credible set.

    Returns
    -------
    int
        Number of variants to include in the credible set.
    """
    return np.sum(np.cumsum(np.sort(alpha)[::-1]) < coverage) + 1  # type: ignore


def get_purity_x(cs, R_list, min_abs_cor, n_purity, X_list):
    """
    Calculate the purity of a credible set.

    This function computes the minimum absolute correlation within a credible set across populations.
    If the credible set size exceeds `n_purity`, a random subset of size `n_purity` is used.

    Parameters
    ----------
    cs : list
        List of indices representing the credible set.
    R_list : list of np.ndarray
        List of LD correlation matrices for each population.
    min_abs_cor : float
        Minimum absolute correlation threshold.
    n_purity : int
        Maximum number of variants to consider for purity calculation.
    X_list : list of np.ndarray
        List of genotype matrices for each population.

    Returns
    -------
    float
        Minimum absolute correlation within the credible set across populations.
    """
    if len(cs) > n_purity:
        cs = random.sample(cs.tolist(), n_purity)
    if R_list is None:
        R_list = [
            np.corrcoef(X_list[i][:, cs], rowvar=False) for i in range(len(X_list))
        ]
    else:
        R_list = [R[cs, :][:, cs] for R in R_list]
    abs_meta_R = functools.reduce(np.maximum, [np.abs(R) for R in R_list])

    return np.min(abs_meta_R)


def recover_R_from_XTX(XTX: np.ndarray, X_l2: np.ndarray) -> None:
    """
    Recover the correlation matrix R from the XTX matrix.

    This function modifies the input XTX matrix in place to recover the correlation matrix R.
    It assumes that the diagonal elements of XTX represent the variances of the variables.

    Parameters
    ----------
    XTX : np.ndarray
        The input XTX matrix (PxP) representing the cross-product of the design matrix.
    X_l2 : np.ndarray
        The diagonal elements of XTX representing the variances of the variables.

    Returns
    -------
    None
    """
    assert (XTX[:, np.flatnonzero(X_l2 == 0)] == 0).all()
    assert (XTX[np.flatnonzero(X_l2 == 0), :] == 0).all()
    with np.errstate(divide="ignore", invalid="ignore"):
        XTX /= np.sqrt(np.nan_to_num(X_l2, 1))  # type: ignore
        XTX /= np.sqrt(np.expand_dims(np.nan_to_num(X_l2, 1), 1))  # type: ignore
